
import unittest

from soapcw import cw
import numpy as np



class TestCWLoad(unittest.TestCase):
    """Tests for cw loading part of the SOAP package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.sig = cw.GenerateSignal(alpha=0.5, delta=0.5, psi=0,phi0=0.0,cosi=0.1,h0=1e-15,f=[200, 0.0])

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_edat_load(self):
        self.sig.get_edat()

    def test_detector_velocity(self):
        self.sig.get_edat()
        detv = self.sig.detector_velocity(self.sig.edat, 1234567987, "H1")
        np.testing.assert_almost_equal(detv , np.array([-0.00005208568783271258, -0.00007807592362349714, -0.00003427742373117495]))

    def test_get_detector_velocities(self):
        self.sig.get_detector_velocities([1234567987], "H1")
        np.testing.assert_almost_equal(self.sig.det_vels[0] , np.array([-0.00005208568783271258, -0.00007807592362349714, -0.00003427742373117495]))

    def test_pulsar_path(self):
        self.assertAlmostEqual(199.98212067259186, self.sig.get_pulsar_path(np.array([1234567987]),"H1")[0])


import os
import shutil
import h5py
import json
import numpy as np
from collections import OrderedDict
import importlib.resources as pkg_resources
import importlib_resources
import pandas as pd

def make_directory_structure(root_dir):

    my_resources = importlib_resources.files("pipeline")
    cssfile = (my_resources / "css"/ "general.css")
    javascriptfile = (my_resources / "scripts"/ "table_scripts.js")
    
    if not os.path.isdir(os.path.join(root_dir, "css")):
        os.makedirs(os.path.join(root_dir, "css"))

    if not os.path.isdir(os.path.join(root_dir, "scripts")):
        os.makedirs(os.path.join(root_dir, "scripts"))

    shutil.copy(cssfile, os.path.join(root_dir, "css/"))
    shutil.copy(javascriptfile, os.path.join(root_dir, "scripts/"))
    """
    shutil.copy("../css/general.css", os.path.join(root_dir, "css/"))
    shutil.copy("../scripts/table_scripts.js", os.path.join(root_dir, "scripts/"))
    """

def create_page_usage_page():
    html = ''' <!DOCTYPE html>
    <html lang="en">
    <head>
    <title>Usage</title>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="stylesheet" href="css/general.css" media="screen" type="text/css">
    </head>
    <body>

    <div class="navbar">
    <a href="index.html" class="active">Home</a>
    <a href="astrophysical/astropage.html">Astrophysical Search</a>
    <a href="lines/linepage.html">Line Search</a>
    <a href=""https://git.ligo.org/joseph.bayley/soapcw"" class="right">Code</a>
    <a href="page_usage.html" class="right">About/Usage</a>
    </div>

    <div class="row">
    <div class="side">
        <h2>SOAP</h2>
        <div><img style="width:50%;margin-right:auto;margin-left:auto;display:block" src="https://git.ligo.org/joseph.bayley/soapcw/-/raw/master/logo/drawing.png" alt="SOAP logo"/></div>
        <p>The SOAP search is a rapid search for continuous gravitational waves using the Viterbi algorithm</p>
        <h3>Links</h3>
        <p>GitLab: <a href="https://git.ligo.org/joseph.bayley/soapcw">https://git.ligo.org/joseph.bayley/soapcw</a></p>
        <p><a href="https://joseph.bayley.docs.ligo.org/soapcw/index.html">Documentation</a></p>

        <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.100.023006">Methods Paper</a></p>
        <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.102.083024">CNN followup paper</a></p>
        <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.106.083022">Parameter estimation paper</a></p>
    </div>
    <div class="main">
        <h2>How to use these pages</h2>
        The run pages are built around a table of each of the frequency bands that are run on and an image showing the output spectrograms and viterbi tracks along with other information.
    
        To filter which frequency bands to use the box to the top left gives a set of options

        <img style="width:30%;margin-right:auto;margin-left:auto;display:block" src="https://git.ligo.org/joseph.bayley/soapcw/-/raw/master/pipeline/images/options.png" />
        <ul>
        <li> In the first box one can sort the table based on one of the selected headings in ascending or descending order. </li>
        <li> In the second box there is a choice of a number of variables, checking these will include them in the table. </li>
        <li> The third box filters the frequency ranges, input the ranges and clikc filter to change the frequency range </li>
        <li> The fourth box has two buttons (last page, next page), these move between the table pages forwards and backwards </li>
        <ul>

        <img style="width:30%;margin-right:auto;margin-left:auto;display:block" src="https://git.ligo.org/joseph.bayley/soapcw/-/raw/master/pipeline/images/table.png" />

        The table then shows a list of different frequency bands defined by the fmin and fmax.
        By default it shows the lineaware stat, which is the line aware statistic output from the SOAP search, and the info column. 
        The info column contains extra information about which is known about that band, i.e. info from line investigations or hw injections

        By clicking on one of these cells it shows a plot and a table containing all of the information about this band.
        The table of information looks like this:

        <img style="width:100%;margin-right:auto;margin-left:auto;display:block" src="https://git.ligo.org/joseph.bayley/soapcw/-/raw/master/pipeline/images/band_summary.png" />

        Which contains all table columns. The hide track button is used to toggle the viterbi track in the plot. 
        The previous and next buttons cycle through the table in order.

        <img style="width:60%;margin-right:auto;margin-left:auto;display:block" src="https://git.ligo.org/joseph.bayley/soapcw/-/raw/master/pipeline/images/plot.png" />

        The plot then contains a number of panels showing information about the band.
        <ul>
        <li> The first panel shows the time-frequency spectrogram from L1 with the viterbi track overlayed </li>
        <li> The second panel shows the time-frequency spectrogram from H1 with the viterbi track overlayed </li>
        <li> The third panel shows the viterbi map from the combination of both detectors, this is related to the probability that a signal is present in any particular time-frequency bin </li>
        <li> The fourth panel shows the normalised SFT power along the red Viterbi track </li>
        <li> The fifth panel shows the mean noise floor across the frequency band as a function of time. </li>
        <ul>
    </div>
    </div>

    <div class="footer">
        <p>If you have any problems with this page please submit an issue to: https://git.ligo.org/joseph.bayley/soapcw/-/issues or contact: joseph.bayley@glasgow.ac.uk </p>

    </div>

    </body>
    </html>'''
    return html


def create_home_page():


    html = ''' <!DOCTYPE html>
    <html lang="en">
    <head>
    <title>SOAP</title>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="stylesheet" href="css/general.css" media="screen" type="text/css">
    </head>
    <body>

    <div class="navbar">
    <a href="index.html" class="active">Home</a>
    <a href="astrophysical/astropage.html">Astrophysical Search</a>
    <a href="lines/linepage.html">Line Search</a>
    <a href="https://git.ligo.org/joseph.bayley/soapcw" class="right">Code</a>
    <a href="page_usage.html" class="right">About/Usage</a>
    </div>

    <div class="row">
    <div class="side">
        <h2>SOAP</h2>
        <div><img style="width:50%;margin-right:auto;margin-left:auto;display:block" src="https://git.ligo.org/joseph.bayley/soapcw/-/raw/master/logo/drawing.png" alt="SOAP logo"/></div>
        <p>The SOAP search is a rapid search for continuous gravitational waves using the Viterbi algorithm</p>
        <h3>Links</h3>
        <p>GitLab: <a href="https://git.ligo.org/joseph.bayley/soapcw">https://git.ligo.org/joseph.bayley/soapcw</a></p>
        <p><a href="https://joseph.bayley.docs.ligo.org/soapcw/index.html">Documentation</a></p>

        <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.100.023006">Methods Paper</a></p>
        <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.102.083024">CNN followup paper</a></p>
        <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.106.083022">Parameter estimation paper</a></p>
    </div>
    <div class="main">
        <h2>General info about SOAP</h2>
        <img style="width:70%" src="https://git.ligo.org/joseph.bayley/soapcw/-/raw/master/pipeline/images/vitmap_ex.png" />
        <br>
        <h2>Astrophysical search</h2>
        <p>The astrophysical search runs using sets of 1800s SFTs summed over one day and combining each detector using the <a href="https://joseph.bayley.docs.ligo.org/soapcw/bayesianlineaware.html">line aware statistic</a></p>
        <h2>Line search</h2>
        <p>The line search also runs on sets of 1800s SFTs summed over one day, however are the search is run separately for each detector to identify instrumental artefacts within the data.</p>


    </div>
    </div>

    <div class="footer">
        <p>If you have any problems with this page please submit an issue to: https://git.ligo.org/joseph.bayley/soapcw/-/issues or contact: joseph.bayley@glasgow.ac.uk </p>

    </div>

    </body>
    </html>'''
    return html

def create_astro_page(run_headings, sub_headings):

    html = f'''<!DOCTYPE html>
            <html lang="en">
            <head>
            <title>Astro Page</title>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <link rel="stylesheet" href="../css/general.css" media="screen" type="text/css">
            <script src="https://ajax.googleapis.com/ajax/libs/jquery/1.7.1/jquery.min.js"></script>
            </head>
            <body>


            <div class="navbar">
            <a href="../index.html" class="active">Home</a>
            <a href="../astrophysical/astropage.html">Astrophysical Search</a>
            <a href="../lines/linepage.html">Line Search</a>

            <a></a>
            {run_headings}

            <a href="https://git.ligo.org/joseph.bayley/soapcw" class="right">Code</a>
            <a href="../page_usage.html" class="right">About/Usage</a>
            </div>


            <div class="row">
            <div class="side">
                <h2>About</h2>
                <div><img style="width:50%;margin-right:auto;margin-left:auto;display:block" src="https://git.ligo.org/joseph.bayley/soapcw/-/raw/master/logo/drawing.png" alt="SOAP logo"/></div>
                    <p>The SOAP search is a rapid search for continuous gravitational waves using the Viterbi algorithm</p>
                    <h3>Links</h3>
                    <p>GitLab: <a href="https://git.ligo.org/joseph.bayley/soapcw">https://git.ligo.org/joseph.bayley/soapcw</a></p>
                    <p><a href="https://joseph.bayley.docs.ligo.org/soapcw/index.html">Documentation</a></p>

                    <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.100.023006">Methods Paper</a></p>
                    <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.102.083024">CNN followup paper</a></p>
                    <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.106.083022">Parameter estimation paper</a></p>
            </div>
            <div class="main">
                <h2>List of runs with information</h2>
                <p>For each of the observing runs there were multiple runs of SOAP with different settings, the recommended one to use is at the top of the list.</p>
                {sub_headings}
            </div>
            </div>

            <div class="footer">
                <p>If you have any problems with this page please submit an issue to: https://git.ligo.org/joseph.bayley/soapcw/-/issues or contact: joseph.bayley@glasgow.ac.uk </p>
            </div>

            </body>

            </html>'''

    return html

def create_line_page(run_headings, sub_headings):

    html = f'''<!DOCTYPE html>
            <html lang="en">
            <head>
            <title>Line Page</title>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <link rel="stylesheet" href="../css/general.css" media="screen" type="text/css">
            <script src="https://ajax.googleapis.com/ajax/libs/jquery/1.7.1/jquery.min.js"></script>
            </head>
            <body>


            <div class="navbar">
            <a href="../index.html" class="active">Home</a>
            <a href="../astrophysical/astropage.html">Astrophysical Search</a>
            <a href="../lines/linepage.html">Line Search</a>

            <a></a>
            {run_headings}

            <a href="https://git.ligo.org/joseph.bayley/soapcw" class="right">Code</a>
            <a href="../page_usage.html" class="right">About/Usage</a>
            </div>


            <div class="row">
            <div class="side">
                <h2>About</h2>
                <div><img style="width:50%;margin-right:auto;margin-left:auto;display:block" src="https://git.ligo.org/joseph.bayley/soapcw/-/raw/master/logo/drawing.png" alt="SOAP logo"/></div>
                    <p>The SOAP search is a rapid search for continuous gravitational waves using the Viterbi algorithm</p>
                    <h3>Links</h3>
                    <p>GitLab: <a href="https://git.ligo.org/joseph.bayley/soapcw">https://git.ligo.org/joseph.bayley/soapcw</a></p>
                    <p><a href="https://joseph.bayley.docs.ligo.org/soapcw/index.html">Documentation</a></p>

                    <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.100.023006">Methods Paper</a></p>
                    <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.102.083024">CNN followup paper</a></p>
                    <p><a href="https://journals.aps.org/prd/abstract/10.1103/PhysRevD.106.083022">Parameter estimation paper</a></p>
            </div>
            <div class="main">
                <h2>List of runs with information</h2>
                <p>SOAP is run separately for each detector, without the Baysian line aware statistic, using just the summed power as a statistic.</p>
                {sub_headings}
            </div>
            </div>

            <div class="footer">
                <p>If you have any problems with this page please submit an issue to: https://git.ligo.org/joseph.bayley/soapcw/-/issues or contact: joseph.bayley@glasgow.ac.uk </p>
            </div>

            </body>

            </html>'''

    return html



def create_run_page(run_headings, obs_run="run"):


    html = f'''
    <!DOCTYPE html>
    <html lang="en">
    <head>
    <title>{obs_run}</title>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="stylesheet" href="../../../css/general.css" media="screen" type="text/css">

    <link href="https://maxcdn.bootstrapcdn.com/bootstrap/3.2.0/css/bootstrap.min.css" rel="stylesheet">   
    <script src="https://ajax.googleapis.com/ajax/libs/jquery/1.7.1/jquery.min.js"></script>
    <link rel="stylesheet" href="https://cdn.datatables.net/1.10.2/css/jquery.dataTables.min.css"></style>
    <script type="text/javascript" src="https://cdn.datatables.net/1.10.2/js/jquery.dataTables.min.js"></script>
    <script type="text/javascript" src="https://maxcdn.bootstrapcdn.com/bootstrap/3.2.0/js/bootstrap.min.js"></script>


    </head>

    <body>

    <div class="navbar">
    <a href="../../../index.html" class="active">Home</a>
    <a href="../../../astrophysical/astropage.html">Astrophysical Search</a>
    <a href="../../../lines/linepage.html">Line Search</a>
    <a></a>
    {run_headings}
    <a href="https://git.ligo.org/joseph.bayley/soapcw" class="right">Code</a>
    <a href="../../../page_usage.html" class="right">About/Usage</a>
    </div>

    <div class="row">
    <div class="side"> 

        <div style="display:flex">
            <button type="button" class="btn btn-cl" style="width:100%; margin:3px" id="filterdivbutton" onclick="filterDivs(event)"> Hide filters</button>
        </div>
        <div class="box" id="sortdiv">
        Sort table by: 
        <select name="Sortby" id="sortby" key="fmin" onchange=reloadPage(event)></select>
        <select name="Sortorder" id="sortorder" key="descending" onchange=reloadPage(event)>
            <option key="ascending">Ascending</option>
            <option key="descending">Descending</option>
        </select>
        </div>

        <div class="box" id="checkboxdiv">
        Include column in table:
        <form id="checkboxform">
        </form>
        </div>

        <div class="box" id="filterdiv">
        Filter frequencies: 
        </br>
        Min: <input type="text" id="minfreq" name="minfreq" size="4" maxlength="6" value="0"> Hz ----
        Max: <input type="text" id="maxfreq" name="maxfreq" size="4" maxlength="6" value="2000"> Hz 
        </br>
        <button type="button" style="margin:3px" class="btn btn-cl py-3" id="filtertable" onclick="reloadPage(event)"> Filter</button>
        <button type="button" class="btn btn-cl py-3" id="filtertablereset" onclick="resetTable(event)"> Reset</button>
        </div>
        <div class="box" id="filterlinediv">
        Filter Info: 
        </br>
        <input type="checkbox" id="onlyhwinjs" name="onlyhwinjs" value="1" onclick=reloadPage(event)> Only Hw injs</input>
        <input type="checkbox" id="onlyknownlines" name="onlyknownlines" value="1" onclick=reloadPage(event)> Only known lines</input>
        </br>
        <input type="checkbox" id="hidehwinjs" name="hidehwinjs" value="1" onclick=reloadPage(event)> Hide Hw injs</input>
        <input type="checkbox" id="hideknownlines" name="hideknownlines" value="1" onclick=reloadPage(event)> Hide known lines</input>
        </div>
        <div class="box" id="showalldiv">
        <button type="button" class="btn btn-cl py-3" id="lastpagebutton" value=0 onclick="lastTablePage(event)"> Last page</button>
        <button type="button" class="btn btn-cl py-3" id="nextpagebutton" value=0 onclick="nextTablePage(event)"> Next page</button>
        </div>

        <table cellpadding="4" border="1" class="table table-bordered table-striped" id="inTable" > 
        </table>
    </div>

    <div class="main">
        <div> 
        <button type="button" style="margin:3px" class="btn btn-cl" id="trackbutton" value="0" onclick="showHideTrack(event)">Hide track</button> 
        <button type="button" style="margin:3px" class="btn btn-cl py-3" id="prevbutton" value="0" onclick="previousImage(event)">Previous</button>
        <button type="button" style="margin:3px" class="btn btn-cl py-3" id="nextbutton" value="0" onclick="nextImage(event)">Next</button> 
        
        <table cellpadding="4" border="1" id="infoTable" class="table table-striped table-bordered"> 
        </table>
        </div>

        <div class="imagewrap"> 
        <img src="" id="image" value="0"></img> 
        </div>
    </div>
    </div>

    <div class="footer">
    <p>If you have any problems with this page please submit an issue to: https://git.ligo.org/joseph.bayley/soapcw/-/issues or contact: joseph.bayley@glasgow.ac.uk </p>
    </div>

    <script type="text/javascript" src="../../../scripts/table_scripts.js"></script>

    </body>


    </html>
    '''

    # button to be added to show all of the frequency bands
    """ <div class="box" id="showalldiv">
        <button type="button" class="btn btn-cl py-3" id="showallbutton" onclick="loadAllBands(event)"> Load all bands</button>
        </div>"""

    return html

def create_line_run_page(run_headings, obs_run="run"):


    html = f'''
    <!DOCTYPE html>
    <html lang="en">
    <head>
    <title>{obs_run}</title>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="stylesheet" href="../../../../css/general.css" media="screen" type="text/css">

    <link href="https://maxcdn.bootstrapcdn.com/bootstrap/3.2.0/css/bootstrap.min.css" rel="stylesheet">   
    <script src="https://ajax.googleapis.com/ajax/libs/jquery/1.7.1/jquery.min.js"></script>
    <link rel="stylesheet" href="https://cdn.datatables.net/1.10.2/css/jquery.dataTables.min.css"></style>
    <script type="text/javascript" src="https://cdn.datatables.net/1.10.2/js/jquery.dataTables.min.js"></script>
    <script type="text/javascript" src="https://maxcdn.bootstrapcdn.com/bootstrap/3.2.0/js/bootstrap.min.js"></script>


    </head>

    <body>

    <div class="navbar">
    <a href="../../../index.html" class="active">Home</a>
    <a href="../../../astrophysical/astropage.html">Astrophysical Search</a>
    <a href="../../../lines/linepage.html">Line Search</a>
    <a></a>
    {run_headings}
    <a href="https://git.ligo.org/joseph.bayley/soapcw" class="right">Code</a>
    <a href="../../../page_usage.html" class="right">About/Usage</a>
    </div>

    <div class="row">
    <div class="side"> 

        <div style="display:flex">
            <button type="button" class="btn btn-cl" style="width:100%; margin:3px" id="filterdivbutton" onclick="filterDivs(event)"> Hide filters</button>
        </div>
        <div class="box" id="sortdiv">
        Sort table by: 
        <select name="Sortby" id="sortby" key="fmin" onchange=reloadPage(event)></select>
        <select name="Sortorder" id="sortorder" key="descending" onchange=reloadPage(event)>
            <option key="ascending">Ascending</option>
            <option key="descending">Descending</option>
        </select>
        </div>

        <div class="box" id="checkboxdiv">
        Include column in table:
        <form id="checkboxform">
        </form>
        </div>

        <div class="box" id="filterdiv">
        Filter frequencies: 
        </br>
        Min: <input type="text" id="minfreq" name="minfreq" size="4" maxlength="6" value="0"> Hz ----
        Max: <input type="text" id="maxfreq" name="maxfreq" size="4" maxlength="6" value="2000"> Hz 
        </br>
        <button type="button" style="margin:3px" class="btn btn-cl py-3" id="filtertable" onclick="reloadPage(event)"> Filter</button>
        <button type="button" class="btn btn-cl py-3" id="filtertablereset" onclick="resetTable(event)"> Reset</button>
        </div>
        <div class="box" id="showalldiv">
        <button type="button" class="btn btn-cl py-3" id="lastpagebutton" value=0 onclick="lastTablePage(event)"> Last page</button>
        <button type="button" class="btn btn-cl py-3" id="nextpagebutton" value=0 onclick="nextTablePage(event)"> Next page</button>
        </div>

        <table cellpadding="4" border="1" class="table table-bordered table-striped" id="inTable" > 
        </table>
    </div>

    <div class="main">
        <div> 
        <button type="button" style="margin:3px" class="btn btn-cl" id="trackbutton" value="0" onclick="showHideTrack(event)">Hide track</button> 
        <button type="button" style="margin:3px" class="btn btn-cl py-3" id="prevbutton" value="0" onclick="previousImage(event)">Previous</button>
        <button type="button" style="margin:3px" class="btn btn-cl py-3" id="nextbutton" value="0" onclick="nextImage(event)">Next</button> 
        
        <table cellpadding="4" border="1" id="infoTable" class="table table-striped table-bordered"> 
        </table>
        </div>

        <div class="imagewrap"> 
        <img src="" id="image" value="0"></img> 
        </div>
    </div>
    </div>

    <div class="footer">
    <p>If you have any problems with this page please submit an issue to: https://git.ligo.org/joseph.bayley/soapcw/-/issues or contact: joseph.bayley@glasgow.ac.uk </p>
    </div>

    <script type="text/javascript" src="../../../scripts/table_scripts.js"></script>

    </body>


    </html>
    '''

    # button to be added to show all of the frequency bands
    """ <div class="box" id="showalldiv">
        <button type="button" class="btn btn-cl py-3" id="showallbutton" onclick="loadAllBands(event)"> Load all bands</button>
        </div>"""

    return html


def read_line_files_old(linefile,det=None):
    """
    open line list file and save information on line
    returns
    --------
    linelist: list
        list of lines []
    """
    data = []
    with open(linefile,"r") as f: 
        i = 0
        for line in f.readlines(): 
            if i == 0:
                i +=1
                continue
            if not line.startswith("%"): 
                lnsplit = line.split("\t")
                lnsave = []
                for ln in lnsplit:
                    try:
                        lnsave.append(float(ln))
                    except:
                        lnsave.append("{} {} {} \n".format(det,lnsplit[0],ln))
                data.append(lnsave)
                del lnsplit,lnsave

    return data

def get_line_info_old(linedata, flow, fhigh):
    info = ""
    # if line files have been loaded include information on known lines
    if linedata is not None:
        for line in linedata:
            # if type of line is not a comb
            if line[1] == 0:
                lowfreq = line[0] - line[5]
                highfreq = line[0] + line[6]

                if flow < lowfreq < fhigh or flow < highfreq < fhigh or flow < line[0] < fhigh:
                    info += line[7]
            # if type of line is a comb include the initial frequency, and first harmonic
            elif line[1] == 1:
                spacing = line[0]
                fharm = line[3]
                lharm = line[4]
                offset = line[2]

                ranges = np.arange(fharm,lharm)*spacing + offset
                for comb in ranges:
                    if flow < comb < fhigh:
                        info += line[7]

    if info == "" or info == np.nan or info == "nan" or info == "NaN":
        pass
    else:
        info = "lines:" + info
    return info

def get_line_info(linedata, flow, fhigh):
    info = ""
    # if line files have been loaded include information on known lines
    if linedata is not None:

        lines = linedata.loc[linedata["Type (0:line; 1:comb; 2:comb with scaling width)"] == 0]
        lines = lines.loc[
            (lines["Frequency or frequency spacing [Hz]"] + lines[" Left width [Hz]" ] < fhigh) & 
            (lines["Frequency or frequency spacing [Hz]"] - lines[" Right width [Hz]" ] > flow)]

        # the spaces at the front of comments is important for the column name
        for index, line in lines.iterrows():
            info += line[" Comments"]

        combs = linedata.loc[linedata["Type (0:line; 1:comb; 2:comb with scaling width)"] == 1]

        for index, comb in combs.iterrows():
            spacing = comb["Frequency or frequency spacing [Hz]"]
            first = comb["First visible harmonic"]
            last = comb[" Last visible harmonic"]
            offset = comb["Frequency offset [Hz]"]

            comb_freqs = np.arange(first, last, spacing) + offset

            if np.any((comb_freqs < fhigh) & (comb_freqs > flow)):
                info += comb[" Comments"]

    #print(info)
    if info == "" or info == np.nan or info == "nan" or info == "NaN":
        pass
    else:
        info = "lines:" + info
    return info

def get_hwinj_info(hwinjtable, flow, fhigh):
    """ """
    hwinjs = hwinjtable.loc[
            (hwinjtable["f0 (epoch start)"] < fhigh) & 
            (hwinjtable["f0 (epoch start)"] > flow)]

    info = ""
    for index, line in hwinjs.iterrows():
        info += f"<a href='https://ldas-jobs.ligo.caltech.edu/~keith.riles/cw/injections/preO3/preO3_injection_params.html'> hwinj: {line['Pulsar']}</a>"
    
    return info

def make_json_from_hdf5(root_dir, linepaths=None, table_order=None, hwinjfile=None):
    """Loads in all hdf5 files and writes them into json format that can be loaded by javascript into summary pages"""
    hdf5dir = os.path.join(root_dir, "data")

    json_data = []

    if linepaths is not None:
        linedataframes = []
        for linefile in linepaths:
            linedataframes.append(pd.read_csv(linefile))
        linedata = pd.concat(linedataframes, axis=0, ignore_index=True)

    if hwinjfile is not None:
        if hwinjfile.endswith("html"):
            hwinjdata = pd.read_html(hwinjfile, header=0)[0]


    """
    linedata = None
    if linepaths is not None:
        if type(linepaths) == str:
            linedata = read_line_files(linepaths)
        else:
            linedata = []
            for linefile in linepaths:
                if "H1" in linefile:
                    det = "H1"
                if "L1" in linefile:
                    det = "L1"
                linedata.extend(read_line_files(linefile,det = det))
    """

    for fname in os.listdir(hdf5dir):
        with h5py.File(os.path.join(hdf5dir, fname), "r", track_order=True) as f:
            for i in range(len(f[list(f.keys())[0]])):
                temp_data = OrderedDict()
                for key in table_order:
                    if key not in list(f.keys()): continue
                    # convert the plot path to the location on the server (works only for LIGO servers at the moment)
                    if key == "plot_path":
                        path = f[key][i].decode()
                        if "/soap_2/" in path:
                            path = path.replace("/soap_2/","/soap/")
                        path = path.replace("/home/", "https://ldas-jobs.ligo.caltech.edu/~").replace("/public_html","")
                        temp_data[key] = path
                    else:
                        temp_data[key] = np.round(f[key][i],2)
                    #if i > 50:
                    #    sys.exit()
                    info = ""
                    if hwinjfile is not None:
                        info += get_hwinj_info(hwinjdata, f["fmin"][i], f["fmax"][i])
                    if linepaths is not None:
                        info += get_line_info(linedata, f["fmin"][i], f["fmax"][i])
                    temp_data.update({"info":info})
                    temp_data.move_to_end("info")
                json_data.append(temp_data)
       
    
    # sort the table so the highest lineaware statitics shjow first
    try:
        sorted_json_data = sorted(json_data, key=lambda d: d["lineaware_stat"])
    except:
        sorted_json_data = sorted(json_data, key=lambda d: d["fmin"])

    with open(os.path.join(root_dir, "table.json"), "w") as f:
        json.dump(sorted_json_data, f)

    # also save separate list of just the top statistics
    with open(os.path.join(root_dir, "table_toplist.json"), "w") as f:
        json.dump(sorted_json_data[:int(0.1*len(sorted_json_data))], f)

def get_public_dir(root_dir):

    username = root_dir.split("/")[2]
    public = root_dir.split("/public_html/")[1]
    public_html = f"https://ldas-jobs.ligo.caltech.edu/~{username}/{public}"
    return public_html

def get_html_string(root_dir, linepaths=None, table_order=None, force_overwrite=False, hwinjfile=None):

    public_dir = get_public_dir(root_dir)
    print("pbdir: ", public_dir)

    run_headings = ""
    sub_headings = ""

    obsruns = sorted(os.listdir(root_dir))

    for head in obsruns:
        if os.path.isdir(os.path.join(root_dir, head)):
            i = 0
            for subhead in os.listdir(os.path.join(root_dir, head)):
                subdir = os.path.join(root_dir, head, subhead)
                if os.path.isdir(subdir):
                    if i == 0:
                        run_headings += f'<a href="{public_dir}/{head}/{subhead}/{subhead}.html">{head}</a>'
                        i += 1
                    else:
                        continue

    for head in obsruns:
        if os.path.isdir(os.path.join(root_dir, head)):
            sub_headings += f"<h1> {head} </h1> <ul>"
            for subhead in os.listdir(os.path.join(root_dir, head)):
                subdir = os.path.join(root_dir, head, subhead)
                if os.path.isdir(subdir):
                    sub_headings += f'<l1><a href="{public_dir}/{head}/{subhead}/{subhead}.html"> {subhead} </a></li> </br>'

                    if os.path.exists(os.path.join(subdir, "table.json")):
                        if force_overwrite:
                            try:
                                make_json_from_hdf5(subdir, linepaths, table_order, hwinjfile=hwinjfile)
                            except:
                                print(f"WARNING: Cannot recreate json table")
                        else:
                            print(f"WARNING: No new updates to {subhead}, {subdir}")
                    else:
                        make_json_from_hdf5(subdir, linepaths, table_order, hwinjfile=hwinjfile)

                    run_html = create_run_page(run_headings, obs_run=head)
                    with open(os.path.join(subdir, f"{subhead}.html"), "w") as f:
                        f.write(run_html)
            sub_headings += "</ul>"

    return run_headings, sub_headings

def get_html_string_week(root_dir, linepath=None, table_order=None):

    run_headings = ""
    sub_headings = ""
    if os.path.exists(root_dir):
        # head is the observing run
        for head in os.listdir(root_dir):
            if os.path.isdir(os.path.join(root_dir, head)):
                run_headings += f'<a href="./{head}/{head}.html">{head}</a>'
                sub_headings += f"<h1> {head} </h1> <ul>"
                for subhead in os.listdir(os.path.join(root_dir, head)):
                    subdir = os.path.join(root_dir, head, subhead)
                    if os.path.isdir(subdir):
                        for weekrun in os.listdir(subdir):
                            weekdir = os.path.join(line_dir, head, subhead, weekrun)
                            if "week" in weekrun:
                                sub_headings += f'<l1><a href="./{head}/{subhead}/{weekrun}/{subhead}.html"> {subhead} {weekrun}</a></li> </br>'

                                if os.path.exists(os.path.join(weekdir, "table.json")):
                                    try:
                                        make_json_from_hdf5(weekdir, linepaths, table_order)
                                    except:
                                        print(f"WARNING: Cannot recreate json table, no new updates to {subhead}, {weekdir}")
                                else:
                                    make_json_from_hdf5(weekdir, linepaths, table_order)

                                run_html = create_line_run_page(line_run_headings, obs_run=head)
                                with open(os.path.join(weekdir, f"{subhead}.html"), "w") as f:
                                    f.write(run_html)
                            else:
                                sub_headings += f'<l1><a href="./{head}/{subhead}/{subhead}.html"> {subhead} </a></li> </br>'

                                if os.path.exists(os.path.join(subdir, "table.json")):
                                    try:
                                        make_json_from_hdf5(subdir, linepaths, table_order)
                                    except:
                                        print(f"WARNING: Cannot recreate json table, no new updates to {subhead}, {subdir}")
                                else:
                                    make_json_from_hdf5(subdir, linepaths, table_order)

                                run_html = create_line_run_page(line_run_headings, obs_run=head)
                                with open(os.path.join(subdir, f"{subhead}.html"), "w") as f:
                                    f.write(run_html)
                sub_headings += "</ul>"

    return run_headings, sub_headings

def write_pages(root_dir, linepaths, table_order, force_overwrite=False, hwinjfile=None, obs_run="run"):
    """ Generate and write the html pages with the inputs from the directory structure"""
    
    make_directory_structure(root_dir)

    astro_dir = os.path.join(root_dir, "astrophysical")

    run_headings,sub_headings = get_html_string(astro_dir, linepaths=linepaths, table_order=table_order, force_overwrite=force_overwrite, hwinjfile=hwinjfile)

    line_dir = os.path.join(root_dir, "lines")

    #line_run_headings, line_sub_headings = get_html_string_week(line_dir, linepath=linepaths, table_order=table_order)
    line_run_headings,line_sub_headings = get_html_string(line_dir, linepaths=linepaths, table_order=table_order, force_overwrite=force_overwrite, hwinjfile=hwinjfile)

    # create pages
    home_html = create_home_page()
    usage_html = create_page_usage_page()
    astro_html = create_astro_page(run_headings, sub_headings)
    line_html = create_line_page(line_run_headings, line_sub_headings)


    # write pages
    with open(os.path.join(root_dir, "index.html"), "w") as f:
        f.write(home_html)

    with open(os.path.join(root_dir, "page_usage.html"), "w") as f:
        f.write(usage_html)

    with open(os.path.join(astro_dir, "astropage.html"), "w") as f:
        f.write(astro_html)

    with open(os.path.join(line_dir, "linepage.html"), "w") as f:
        f.write(line_html)

 

def main():
    import argparse
    from .soap_config_parser import SOAPConfig
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-c', '--config-file', help='config file', type=str, required=False, default=None)
    parser.add_argument('-o', '--out-path', help='top level of output directories', type=str)
    parser.add_argument('--force-overwrite', help='force overwrite tables', action=argparse.BooleanOptionalAction)
                                                   
    args = parser.parse_args()  

    if args.config_file is not None:
        if not os.path.isfile(args.config_file):
            raise FileNotFoundError(f"No File: {args.config_file}")
        cfg = SOAPConfig(args.config_file)

    else:
        #outpath,minfreq,maxfreq,obs_run="O3",vitmapmodelfname=None, spectmodelfname = None, vitmapstatmodelfname = None, allmodelfname = None, sub_dir = "soap",
        cfg = {"output":{}, "data":{}, "input":{}, }

    if args.out_path:
        cfg["output"]["save_directory"] = args.out_path

    if "lines_h1" in cfg["input"].keys():
        linepaths = [cfg["input"]["lines_h1"], cfg["input"]["lines_l1"]]
    else:
        linepaths = None
    if "hardware_injections" in cfg["input"].keys():
        hwinjfile = cfg["input"]["hardware_injections"]
    else:
        hwinjfile = None

    table_order = ["fmin", "fmax", "lineaware_stat", "H1_viterbistat", "L1_viterbistat", "CNN_vitmap_stat", "CNN_spect_stat", "CNN_vitmapstat_stat", "CNN_all_stat", "plot_path"]

    write_pages(os.path.dirname(os.path.normpath(cfg["output"]["save_directory"])), linepaths, table_order, force_overwrite=args.force_overwrite, hwinjfile=hwinjfile)

if __name__ == "__main__":
    main()


import pkg_resources
import torch
import numpy as np
import configparser
import soapcw

def load_model_from_config(config, weights, device="cpu"):
    """
    Load in model weights from config and weights
    args
    ---------
    config: string
        path to the config ini file
    weights: string
        path to for the checkpointed weights of network

    returns
    ---------
    model_soap: torch.Module
        returns a torch model with trained weights
    config: SoapConfigParser
        returns the parsed config file for the model
    """
    config = soapcw.soap_config_parser.SOAPConfig(config)

    pre_model_weights = torch.load(weights, map_location=device)


    model_soap = soapcw.cnn.pytorch.models.CNN(
        input_dim=config["model"]["img_dim"],               # the size of the input image
        fc_layers=config["model"]["fc_layers"],               # the size of the fully connected mlp layers 
        conv_layers=config["model"]["conv_layers"],           # the convolutional layers (num_filters, filter_size, maxpool_size, stride)
        inchannels=config["model"]["n_channels"], 
        avg_pool_size=config["model"]["avg_pool_size"],                            # the size of the average pooling layer    
        device=device).to(device)


    model_soap.load_state_dict(pre_model_weights["model_state_dict"])

    return model_soap, config
    
def load_vitmap_model(fmin, obs_run = "O3", checkpoint_file = None):
    """
    Load the Viterbimap model for testing
    """
    if 40 < fmin < 500:
        brange = "40_500"
        stride = 1
    elif 500 < fmin < 1000:
        brange = "500_1000"
        stride = 2
    elif 1000 < fmin < 1500:
        brange = "1000_1500"
        stride = 3
    elif 1500 < fmin < 2000:
        brange = "1500_2000"
        stride = 4
        
    if np.round(fmin*10).astype(int) % stride*2 == 0:
        oddeven = "even"
    else:
        oddeven = "odd"
    
    if checkpoint_file is None:
        model_state = pkg_resources.resource_stream(__name__, "trained_models/{}/vitmap_F{}_{}.ckpt".format(obs_run,brange, oddeven))
    model_init = pkg_resources.resource_stream(__name__, "trained_models/{}/vitmap.ini".format(obs_run))
    
    config = read_config(model_init)

    # hardcoded input size at the moment
    model = cnn.CNN((156,89), 1, config["lin_layers"], config["conv_layers"] , device ="cpu", num_channels = 0)
    model.load_state_dict(torch.load(model_state))
    
    return model


def load_spect_model(fmin, obs_run = "O3"):

    if 40 < fmin < 500:
        brange = "40_500"
        stride = 1
    elif 500 < fmin < 1000:
        brange = "500_1000"
        stride = 2
    elif 1000 < fmin < 1500:
        brange = "1000_1500"
        stride = 3
    elif 1500 < fmin < 2000:
        brange = "1500_2000"
        stride = 4
        
    if np.round(fmin*10).astype(int) % stride*2 == 0:
        oddeven = "even"
    else:
        oddeven = "odd"

    model_state = pkg_resources.resource_stream(__name__, "trained_models/{}/vitmap_F{}_{}.ckpt".format(obs_run,brange, oddeven))
    model_init = pkg_resources.resource_stream(__name__, "trained_models/{}/vitmap.ini".format(obs_run))
    
    config = read_config(model_init)

    model = cnn.CNN((156,89), 1, config["lin_layers"], config["conv_layers"] , device ="cpu", num_channels = 2)
    model.load_state_dict(torch.load(model_state))
    
    return model


    
def read_config(config_file):
    cp = configparser.ConfigParser()
    cp.read(config_file.name)

    p = {}
    params = {"model": ["conv_num_filt", "conv_filt_size", "conv_max_pool", "lin_num_neuron", "lin_dropout", "learning_rate", "num_channels"],}

    for key,vals in params.items():
        for val in vals:
            try:
                p[val] = cp.get(key,val)
            except:
                p[val] = None
                print("No key: {}, {}".format(key,val))

    p["lin_layers"] = [int(n) for n in p["lin_num_neuron"].split(" ")]
    conv_num_filt = [int(n) for n in p["conv_num_filt"].split(" ")]
    conv_filt_size = [int(n) for n in p["conv_filt_size"].split(" ")]
    conv_max_pool = [int(n) for n in p["conv_max_pool"].split(" ")]
    p["conv_layers"] = [(conv_num_filt[i], conv_filt_size[i], conv_max_pool[i], 1) for i in range(len(conv_num_filt))]

    return p




from astropy.coordinates import SkyCoord
import numpy as np

def transform_ecliptic_geo(lon, lat):
    """ transform ecliptic coords to geocentric coords with astropy"""
    skyp = SkyCoord(lon = lon,lat=lat,unit="rad",frame="barycentrictrueecliptic")
    skyp_icrs = skyp.icrs
    alpha, delta = skyp_icrs.data.lon.rad,skyp_icrs.data.lat.rad
    return alpha, delta

def transform_normlonlat_lonlat(fmin, width, lon, lat, f, fdot):
    """_summary_                                                                                                        
    Args:                                                                                                                                                            
        fmin (_type_): _description_             
        width (_type_): _description_                                                                                                                                
        lon (_type_): _description_                     
        lat (_type_): _description_            
        f (_type_): _description_                  
        fdot (_type_): _description_      
    Returns:                      
        _type_: _description_                    
    """
    lon = lon*2*np.pi
    lon = np.remainder(lon, 2*np.pi)
    lat = np.abs(lat)*(0.5*np.pi)

    f = f*width + fmin
    fdot = (fdot*2 - 1)*1e-9

    return lon, lat, f, fdot

def transform_normlonlat_alphadelta(fmin, width, lon, lat, f, fdot):
    """_summary_

    Args:
        fmin (_type_): _description_
        width (_type_): _description_
        lon (_type_): _description_
        lat (_type_): _description_
        f (_type_): _description_
        fdot (_type_): _description_

    Returns:
        _type_: _description_
    """

    
    lon, lat, f, fdot = transform_normlonlat_lonlat(fmin, width, lon, lat, f, fdot)

    alpha, delta = transform_ecliptic_geo(lon, lat)

    return alpha, delta, f, fdot



import sys
import os
import appdirs
import requests

this_dir, this_filename = os.path.split(__file__)
DATA_PATH = os.path.join(this_dir, "data", "data.txt")

LAL_EPHEMERIS_URL = "https://git.ligo.org/lscsoft/lalsuite/raw/master/lalpulsar/lib/{}"
EPHEMERIS_CACHE_DIR = appdirs.user_cache_dir(appname="soapcw", appauthor=False)

def download_ephemeris_file(url):
    """
    !!! This is taken from Matt pitkins cwinpy package https://git.ligo.org/cwinpy/cwinpy/-/blob/master/cwinpy/utils.py !!!
    Download and cache an ephemeris files from a given URL. If the file has
    already been downloaded and cached it will just be retrieved from the cache
    location.

    Parameters
    ----------
    url: str
        The URL of the file to download.
    """

    fname = os.path.basename(url)  # extract the file name
    fpath = os.path.join(EPHEMERIS_CACHE_DIR, fname)

    if os.path.isfile(fpath):
        # return previously cached file
        return fpath

    # try downloading the file
    try:
        ephdata = requests.get(url)
    except Exception as e:
        raise RuntimeError(f"Error downloading from {url}\n{e}")

    if ephdata.status_code != 200:
        raise RuntimeError(f"Error downloading from {url}")

    if not os.path.exists(EPHEMERIS_CACHE_DIR):
        try:
            os.makedirs(EPHEMERIS_CACHE_DIR)
        except OSError:
            if not os.path.exists(EPHEMERIS_CACHE_DIR):
                raise
    elif not os.path.isdir(EPHEMERIS_CACHE_DIR):
        raise OSError(f"Cache directory {EPHEMERIS_CACHE_DIR} is not a directory")

    # write out file to cache
    with open(fpath, "wb") as fp:
        fp.write(ephdata.content)

    return fpath





import math
from numbers import Number
import torch
from torch.distributions import Distribution, constraints
from torch.distributions.utils import broadcast_all

CONST_SQRT_2 = math.sqrt(2)
CONST_INV_SQRT_2PI = 1 / math.sqrt(2 * math.pi)
CONST_INV_SQRT_2 = 1 / math.sqrt(2)
CONST_LOG_INV_SQRT_2PI = math.log(CONST_INV_SQRT_2PI)
CONST_LOG_SQRT_2PI_E = 0.5 * math.log(2 * math.pi * math.e)


class TruncatedStandardNormal(Distribution):
    """
    Truncated Standard Normal distribution
    https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    """

    arg_constraints = {
        'a': constraints.real,
        'b': constraints.real,
    }
    has_rsample = True

    def __init__(self, a, b, validate_args=None):
        self.a, self.b = broadcast_all(a, b)
        if isinstance(a, Number) and isinstance(b, Number):
            batch_shape = torch.Size()
        else:
            batch_shape = self.a.size()
        super(TruncatedStandardNormal, self).__init__(batch_shape, validate_args=validate_args)
        if self.a.dtype != self.b.dtype:
            raise ValueError('Truncation bounds types are different')
        if any((self.a >= self.b).view(-1,).tolist()):
            raise ValueError('Incorrect truncation range')
        eps = torch.finfo(self.a.dtype).eps
        self._dtype_min_gt_0 = eps
        self._dtype_max_lt_1 = 1 - eps
        self._little_phi_a = self._little_phi(self.a)
        self._little_phi_b = self._little_phi(self.b)
        self._big_phi_a = self._big_phi(self.a)
        self._big_phi_b = self._big_phi(self.b)
        self._Z = (self._big_phi_b - self._big_phi_a).clamp_min(eps)
        self._log_Z = self._Z.log()
        little_phi_coeff_a = torch.nan_to_num(self.a, nan=math.nan)
        little_phi_coeff_b = torch.nan_to_num(self.b, nan=math.nan)
        self._lpbb_m_lpaa_d_Z = (self._little_phi_b * little_phi_coeff_b - self._little_phi_a * little_phi_coeff_a) / self._Z
        self._mean = -(self._little_phi_b - self._little_phi_a) / self._Z
        self._variance = 1 - self._lpbb_m_lpaa_d_Z - ((self._little_phi_b - self._little_phi_a) / self._Z) ** 2
        self._entropy = CONST_LOG_SQRT_2PI_E + self._log_Z - 0.5 * self._lpbb_m_lpaa_d_Z

    @constraints.dependent_property
    def support(self):
        return constraints.interval(self.a, self.b)

    @property
    def mean(self):
        return self._mean

    @property
    def variance(self):
        return self._variance

    @property
    def entropy(self):
        return self._entropy

    @property
    def auc(self):
        return self._Z

    @staticmethod
    def _little_phi(x):
        return (-(x ** 2) * 0.5).exp() * CONST_INV_SQRT_2PI

    @staticmethod
    def _big_phi(x):
        return 0.5 * (1 + (x * CONST_INV_SQRT_2).erf())

    @staticmethod
    def _inv_big_phi(x):
        return CONST_SQRT_2 * (2 * x - 1).erfinv()

    def cdf(self, value):
        if self._validate_args:
            self._validate_sample(value)
        return ((self._big_phi(value) - self._big_phi_a) / self._Z).clamp(0, 1)

    def icdf(self, value):
        return self._inv_big_phi(self._big_phi_a + value * self._Z)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        return CONST_LOG_INV_SQRT_2PI - self._log_Z - (value ** 2) * 0.5

    def rsample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        p = torch.empty(shape, device=self.a.device).uniform_(self._dtype_min_gt_0, self._dtype_max_lt_1)
        return self.icdf(p)


class TruncatedNormal(TruncatedStandardNormal):
    """
    Truncated Normal distribution
    https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    """

    has_rsample = True

    def __init__(self, loc, scale, a, b, validate_args=None):
        self.loc, self.scale, a, b = broadcast_all(loc, scale, a, b)
        a = (a - self.loc) / self.scale
        b = (b - self.loc) / self.scale
        super(TruncatedNormal, self).__init__(a, b, validate_args=validate_args)
        self._log_scale = self.scale.log()
        self._mean = self._mean * self.scale + self.loc
        self._variance = self._variance * self.scale ** 2
        self._entropy += self._log_scale

    def _to_std_rv(self, value):
        return (value - self.loc) / self.scale

    def _from_std_rv(self, value):
        return value * self.scale + self.loc

    def cdf(self, value):
        return super(TruncatedNormal, self).cdf(self._to_std_rv(value))

    def icdf(self, value):
        return self._from_std_rv(super(TruncatedNormal, self).icdf(value))

    def log_prob(self, value):
        return super(TruncatedNormal, self).log_prob(self._to_std_rv(value)) - self._log_scale



try:
    from .import gen_lookup
except:
    from .import gen_lookup_python as gen_lookup
    #print("using python integration for line aware stat (install boost and gsl C++ libraries for faster runtime -- see documentation)")
import numpy as np
import sys
import json
import pickle as pickle
import argparse
import os

def save_lookup_amp(p1,p2,ratio,outdir, ndet=2, pow_range = (1,400,500), frac_range = (0.1,1,10)):
    """
    save the lookup table for two detectors with the line aware statistic with consitistent amplitude
    (uses json to save file)
    Args
    --------------
    p1 : float
        width of prior of signal model
    p2 : float
        width of prior of line model
    ratio : float
        ratio of line to noise models
    outdir: string
        directory to save lookup table file
    pow_range: tuple
        ranges for the spectrogram power (lower, upper, number), default (1,400,500)
    frac_range: tuple
        ranges for the ratios of sensitivity and duty cycle (lower, upper, number), default (0.1,1,10)
    """
    minimum,maximum,num = pow_range
    minn,maxn,numn = frac_range
    #ch_arr_app = gen_data.gen_lookup_noise(np.linspace(minimum,maximum,num),np.linspace(minimum,maximum,num),np.linspace(minn,maxn,numn),int_type="chi2",approx=False,pvs=p1,pvl=p2,beta=ratio)
    filename = outdir+"/ch2_signoiseline_{}det_{}_{}_{}.json".format(ndet,p1,p2,ratio)
    if os.path.isfile(filename):
        print(("File {} exists".format(filename)))
    else:
        ch_arr_app = gen_lookup.LineAwareAmpStatistic(np.linspace(minimum,maximum,num),fraction=np.linspace(minn,maxn,numn), ndet=ndet,signal_prior_width=p1,line_prior_width=p2,noise_line_model_ratio=ratio)
        with open(filename,'w+') as f:
            save_data = [[minimum,maximum,num,minn,maxn,numn],np.log(ch_arr_app[0]).tolist()]

            json.dump(save_data,f)

def save_lookup(p1,p2,ratio,outdir,ndet=2,pow_range = (1,400,500), k=2, N=48):
    """
    save the lookup table for two detectors with the line aware statistic
    
    Args
    --------------
    p1 : float
        width of prior of signal model
    p2 : float
        width of prior of line model
    ratio : float
        ratio of line to noise models
    outdir: string
        directory to save lookup table file
    pow_range: tuple
        ranges for the spectrogram power (lower, upper, number), default (1,400,500)

    """

    minimum,maximum,num = pow_range

    if os.path.isfile(outdir+"/signoiseline_{}det_{}_{}_{}.txt".format(ndet, p1,p2,ratio)):
        pass
    else:
        if ndet == 1:
            ch_arr_app = gen_lookup.LineAwareStatistic(np.linspace(minimum,maximum,num),
                                                        ndet=ndet,
                                                        signal_prior_width=p1,
                                                        line_prior_width=p2,
                                                        noise_line_model_ratio=ratio)
        if ndet == 2:
            powers = np.linspace(minimum,maximum,num)
            ch_arr_app = gen_lookup.LineAwareStatistic(powers=powers,
                                                        ndet=ndet,
                                                        k = k,
                                                        N = N,
                                                        signal_prior_width=p1,
                                                        line_prior_width=p2,
                                                        noise_line_model_ratio=ratio)
        with open(outdir+"/signoiseline_{}det_{}_{}_{}.txt".format(ndet, p1,p2,ratio),'wb') as f:
            header = "{} {} {}".format(minimum,maximum,num)
            np.savetxt(f,np.log(ch_arr_app.signoiseline),header = header)



def resave_files(p1,p2,ratio,output):
    """
    resave text files into pickle format
    """
    if os.path.isfile(output+"/txt/ch2_signoiseline_{}_{}_{}.txt".format(p1,p2,ratio)):
        with open(output+"/txt/ch2_signoiseline_{}_{}_{}.txt".format(p1,p2,ratio),'rb') as f:
            save_array = pickle.load(f)
        if os.path.isdir(output+"/pkl/"):
            pass
        else:
            os.mkdir(output+"/pkl/")
        with open(output+"/pkl/ch2_signoiseline_{}_{}_{}.pkl".format(p1,p2,ratio),'wb') as f:
            pickle.dump(save_array,f,protocol=pickle.HIGHEST_PROTOCOL)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
                    prog = 'SOAP lookup table generation',
                    description = 'generates lookup tables for SOAP',)

    parser.add_argument('--amp-stat',action='store_true') 
    parser.add_argument('-s', '--signal-probability', required=True, type=float) 
    parser.add_argument('-l', '--line-probability', required=True, type=float) 
    parser.add_argument('-n', '--noise-line-ratio', required=True, type=float) 
    parser.add_argument('-ndet', '--ndet', default=2, required=False, type=int) 
    parser.add_argument('-k', default=2, required=False, type=int) 
    parser.add_argument('-N', '--num-sfts', default=48, required=False, type=int) 
    parser.add_argument('-o', '--save-dir', required=True, type=str) 
    
    parser.add_argument('-pmin', '--pow-min', default=1, required=False, type=float) 
    parser.add_argument('-pmax', '--pow-max', default=400, required=False, type=float) 
    parser.add_argument('-np', '--n-powers', default=500, required=False, type=int) 

    parser.add_argument('-fmin', '--frac-min', default=0.1, required=False, type=float) 
    parser.add_argument('-fmax', '--frac-max', default=1, required=False, type=float) 
    parser.add_argument('-nf', '--n-fracs', default=10, required=False, type=int) 

    args = parser.parse_args()

    if not args.amp_stat:
        save_lookup(args.signal_probability,
                    args.line_probability,
                    args.noise_line_ratio,
                    args.save_dir,
                    k = args.k,
                    N = args.num_sfts,
                    ndet=args.ndet,
                    pow_range = (args.pow_min,args.pow_max,args.n_powers))
    else:
        save_lookup_amp(args.signal_probability,
                        args.line_probability,
                        args.noise_line_ratio,
                        args.save_dir, 
                        ndet = args.ndet, 
                        pow_range = (args.pow_min,args.pow_max,args.n_powers), 
                        frac_range = (args.frac_min,args.frac_max,args.n_fracs))



from __future__ import division
import numpy as np
import pickle
from sys import stdout
import scipy.special as sp
try:
    from scipy.misc import logsumexp
except:
    from scipy.special import logsumexp
from libc.math cimport log,exp,sqrt
from libc.stdio cimport printf
import os
#import integrals
cwd = os.getcwd()


cdef extern from "integrals.h":
    double integral_signal_1det(double g, double logfrac, int k, int N, double pv)
    double ch2_noise_1det(double g, int k, int N) 
    double integral_signal_2det(double g1, double g2, double logfrac, int k, int N, double pv)
    double integral_line_2det(double g1, double g2,  double logfrac, int k, int N, double pv)
    double ch2_noise_2det(double g1, double g2, int k, int N) 


class LineAwareStatistic:

    def __init__(self,powers, ndet = 2, k = 2, N = 48, signal_prior_width = 1, line_prior_width = 5, noise_line_model_ratio = 1):

        if ndet == 1:
            self.signoiseline,self.sig,self.line,self.noise = self.gen_lookup_one_det(powers,signal_prior_width=signal_prior_width,line_prior_width=line_prior_width,noise_line_model_ratio=noise_line_model_ratio,k=k,N=N)
        elif ndet == 2:
            print(powers)
            self.signoiseline,self.sig,self.line,self.noise = self.gen_lookup_two_det(powers,powers,signal_prior_width=signal_prior_width,line_prior_width=line_prior_width,noise_line_model_ratio=noise_line_model_ratio,k=k,N=N)
        else:
            raise Exception("This currently only works for 1 or 2 detectors")
    
    def bayes_2det(self,g1,g2,logfrac,k,N,signal_prior_width,line_prior_width,noise_line_model_ratio):
        """
        calculate the bayes factor for signal noise and line
        args
        -----------
        g1: float
            power in det 1
        g2: float
            power in det 2
        logfrac: float
            fraction of duty cycle and noise in each detector
        k: int
            number of degrees of freedom
        N: int
        number of sfts summed over
        signal_prior_width: float
        width of prior on signal
        line_prior_width: float
        width of prior on line
        noise_line_model_ratio: float
        noise_line_model_ratio of the prior on noise model vs line
        returns
        ---------
        b1: float
        bayes factor including all signal,line and noise
        sig_int: float
        marginalised likelihood for signal
        line_int:
        marginalised likelihood for line
        noise: float
        noise likelihood
        """
        cdef double sig_int
        cdef double line_int
        cdef double noise
        cdef double b1
        sig_int = integral_signal_2det(g1,g2,logfrac,k,N,signal_prior_width)
        line_int = integral_line_2det(g1,g2,logfrac,k,N,line_prior_width)
        noise = ch2_noise_2det(g1,g2,k,N)
        b1 = sig_int/(noise_line_model_ratio*line_int + noise)
        return [b1,sig_int,line_int,noise]
    
    def bayes_1det(self,g,logfrac,k,N,signal_prior_width,line_prior_width,noise_line_model_ratio):
        """
        calculate the bayes factor for signal noise and line
        args
        -----------
        g1: float
        power in det 1
        g2: float
        power in det 2
        logfrac: float
        fraction of duty cycle and noise in each detector
        k: int
        number of degrees of freedom
        N: int
        number of sfts summed over
        signal_prior_width: float
        width of prior on signal
        line_prior_width: float
        width of prior on line
        noise_line_model_ratio: float
        noise_line_model_ratio of the prior on noise model vs line
        returns
        ---------
        b1: float
        bayes factor including all signal,line and noise
        sig_int: float
        marginalised likelihood for signal
        line_int:
        marginalised likelihood for line
        noise: float
        noise likelihood
        """
        cdef double sig_int
        cdef double line_int
        cdef double noise
        cdef double b1
        sig_int = integral_signal_1det(g,logfrac,k,N,signal_prior_width)
        line_int = integral_signal_1det(g,logfrac,k,N,line_prior_width)
        noise = ch2_noise_1det(g,k,N)
        b1 = sig_int/(noise_line_model_ratio*line_int + noise)
        return [b1,sig_int,line_int,noise]
    
    def gen_lookup_one_det(self,x,signal_prior_width=1,line_prior_width=1,noise_line_model_ratio=1,k=2,N=48):
        """
        calculates the odds ratio for given values of power and ratio.
        args
        --------
        x: array
        array of the SFT power in detector 1
        n1: array
        array of the ratios of the duty cycle and noise psds, \frac{f_{1} S_{2}}{f_{2} S_{1}}
        signal_prior_width: float
        width on prior for signal model
        line_prior_width: float
        width on prior for line model
        noise_line_model_ratio: float
        ratio of the priors on the line and signal models
        k: int
        number of degrees of freedom
        N: int
        number of summed distributions, i.e. SFTs 
        returns
        ----------
        odds: array
        array of odd ratio with 3 models
        signal: array
        array of signal likelihoods
        line: array
        array of line likelihoods
        noise: array
        array of noise likelihoods
        """
        cdef int i
        cdef double[:] xc = x
        cdef double[:] odds = np.zeros((len(x)))
        cdef double[:] signal = np.zeros((len(x)))
        cdef double[:] line = np.zeros((len(x)))
        cdef double[:] noise = np.zeros((len(x)))
        cdef double n1 = 0.
        for i in range(len(x)):
            ig = self.bayes_1det(xc[i],n1,k,N,signal_prior_width,line_prior_width,noise_line_model_ratio)
            odds[i] = ig[0]
            signal[i] = ig[1]
            line[i] = ig[2]
            noise[i] = ig[3]
            
        return np.array(odds),np.array(signal),np.array(line),np.array(noise)


    def gen_lookup_two_det(self,x,y,signal_prior_width=1,line_prior_width=1,noise_line_model_ratio=1,k=2,N=48):
        """
        calculates the odds ratio for given values of power and ratio for two detectors.
        args
        --------
        x: array
        array of the SFT power in detector 1
        y: array
        array of SFT power in detector 2
        n1: array
        array of the ratios of the duty cycle and noise psds, \frac{f_{1} S_{2}}{f_{2} S_{1}}
        int_type: string
        which statistic to use ("chi2" only option for now)
        signal_prior_width: float
        width on prior for signal model
        line_prior_width: float
        width on prior for line model
        noise_line_model_ratio: float
        ratio of the priors on the line and signal models
        k: int
        number of degrees of freedom
        N: int
        number of summed distributions, i.e. SFTs 
        returns
        ----------
        odds: array
        array of odd ratio with 3 models
        signal: array
        array of signal likelihoods
        line: array
        array of line likelihoods
        noise: array
        array of noise likelihoods
        
        """
        
        cdef int i
        cdef int j
        cdef double[:] xc = x
        cdef double[:] yc = y
        cdef double[:, :] odds = np.zeros((len(x),len(y)))
        cdef double[:, :] signal = np.zeros((len(x),len(y)))
        cdef double[:, :] line = np.zeros((len(x),len(y)))
        cdef double[:, :] noise = np.zeros((len(x),len(y)))
        cdef double n1 = 0.
        for i in range(len(x)):
            for j in range(len(y)):
                ig = self.bayes_2det(xc[i],yc[j],n1,k,N,signal_prior_width,line_prior_width,noise_line_model_ratio)
                odds[i,j] = ig[0]
                signal[i,j] = ig[1]
                line[i,j] = ig[2]
                noise[i,j] = ig[3]
                
        return np.array(odds),np.array(signal),np.array(line),np.array(noise)


class LineAwareAmpStatistic(LineAwareStatistic):

    def __init__(self,powers, fractions=None, ndet = 2, k = 2, N = 48, signal_prior_width = 1, line_prior_width = 5, noise_line_model_ratio = 1):

        self.ndet = ndet
        self.signal_prior_width = signal_prior_width
        self.line_prior_width = line_prior_width
        self.noise_line_model_ratio = noise_line_model_ratio
        self.k = k
        self.N = N
        
        if ndet == 1:
            self.signoiseline,self.sig,self.line,self.noise = self.gen_lookup_one_det_amp(powers,signal_prior_width=signal_prior_width,line_prior_width=line_prior_width,noise_line_model_ratio=noise_line_model_ratio)
        elif ndet == 2:
            self.signoiseline,self.sig,self.line,self.noise = self.gen_lookup_two_det_amp(powers,powers,fractions,signal_prior_width=signal_prior_width,line_prior_width=line_prior_width,noise_line_model_ratio=noise_line_model_ratio)
        else:
            raise Exception("Currently only works for 1 or two detectors")
            
    def gen_lookup_two_det_amp(self,x,y,n1,signal_prior_width=1,line_prior_width=1,noise_line_model_ratio=1,k=2,N=48):
        """
        calculates the odds ratio in ch_integrals_approx_noise for given values of power and ratio.
        args
        --------
        x: array
        array of the SFT power in detector 1
        y: array
        array of SFT power in detector 2
        n1: array
        array of the ratios of the duty cycle and noise psds, \frac{f_{1} S_{2}}{f_{2} S_{1}} (f is fraction of summed sfts which are real data, S is noise psd)
        int_type: string
        which statistic to use ("chi2" only option for now)
        approx: bool
        choose whether to use trapz or quad to complete integrals
        signal_prior_width: float
        width on prior for signal model
        line_prior_width: float
        width on prior for line model
        noise_line_model_ratio: float
        ratio of the priors on the line and signal models
        k: int
        number of degrees of freedom for chi2 dist (default = 2)
        N: int
        number of summed sfts (default = 48)
        returns
        ----------
        val: array
        array of odd ratio with 3 models
        val1: array
        array of odd ratio with signal/noise model
        val2: array
        array of odd ratio with signal/line model
        val3: array
        array of signal likelihoods
        val4: array
        array of noise likelihoods
        val: array
        array of line likelihoods
        """
        cdef int i
        cdef int j
        cdef int l
        cdef double[:] xc = x
        cdef double[:] yc = y
        cdef double[:] n1c = n1
        cdef double[:, :, :] odds = np.zeros((len(n1),len(x),len(y)))
        cdef double[:, :, :] signal = np.zeros((len(n1),len(x),len(y)))
        cdef double[:, :, :] line = np.zeros((len(n1),len(x),len(y)))
        cdef double[:, :, :] noise = np.zeros((len(n1),len(x),len(y)))
        #cdef double[:, :, :] val4 = np.zeros((len(x),len(y),len(n1)))
        #cdef double[:, :, :] val5 = np.zeros((len(x),len(y),len(n1)))
        for i in range(len(x)):
            for j in range(len(y)):
                for l in range(len(n1)):
                    ig = self.bayes_2det(xc[i],yc[j],n1c[l],self.k,self.N,signal_prior_width,line_prior_width,noise_line_model_ratio)
                    odds[l,i,j] = ig[0]
                    signal[l,i,j] = ig[1]
                    line[l,i,j] = ig[2]
                    noise[l,i,j] = ig[3]
                    #val4[i,j,k] = ig[4]
                    #val5[i,j,k] = ig[5]

        return np.array(odds),np.array(signal),np.array(line),np.array(noise)#,val4,val5
                

    
    



#!/home/joseph.bayley/.virtualenvs/soap27/bin/python

import sys
import os
import getopt
import time
import configparser 
import random
import numpy as np
import argparse

def create_dirs(dirs):
    for i in dirs:
        if not os.path.isdir(i):
            try: 
                os.makedirs(i)
            except:
                print("Could not create directory {}".format(i))
                sys.exit(1)
    print("All directories exist")


def write_subfile(sub_filename,config_file,config,datatype,comment,dirs, verbose=False):
    log_dir,err_dir,output_dir,condor_dir,save_dir = dirs

    with open(sub_filename,'w') as f:
        f.write('# filename: {}\n'.format(sub_filename))
        f.write('universe = vanilla\n')
        execute = os.path.join(os.path.split(sys.executable)[0],config["code"]["search_exec"])
        f.write('executable = {}\n'.format(execute))
        #f.write('enviroment = ""\n')
        f.write('getenv  = True\n')
        f.write("RequestMemory = 50000 \n")
        f.write("request_disk = 10000 \n")
        f.write('log = {}/{}_$(cluster).log\n'.format(log_dir,comment))
        f.write('error = {}/{}_$(cluster).err\n'.format(err_dir,comment))
        f.write('output = {}/{}_$(cluster).out\n'.format(output_dir,comment))
        args = "arguments = -c {} ".format(config_file)

        args += f' -l $(bandstart) -u $(bandend) -w $(bandwidth) --data-type {datatype} -rt {config["data"]["run"]}'
        
        if config["data"]["run"] == "gauss" and datatype == "train":
            args += f' -np {config["data"]["nperband"]}'
        else:
            args += f' -np 1'

        f.write(args + "\n")
        #f.write('accounting_group = aluk.dev.s6.cw.viterbi\n')
        f.write(f'accounting_group = {config["condor"]["accounting_group"]}\n')
        f.write('queue\n')
    if verbose:
        print(time.time(), "generated subfile")


def write_dagfile(config, datatype="train"):

    condor_dir = "{}/condor".format(config["condor"]["root_dir"])
    output_dir = "{}/condor/output".format(config["condor"]["root_dir"])
    save_dir = "{}".format(config["general"]["save_dir"])
    log_dir = "{}/condor/logs".format(config["condor"]["root_dir"])
    err_dir = "{}/condor/err".format(config["condor"]["root_dir"])
    dirs = [log_dir,err_dir,output_dir,condor_dir,save_dir]
    create_dirs(dirs)

    k = 0

    band_list = []
    for i, bs in enumerate(config["data"]["band_starts"]):
        bandstart = np.arange(config["data"]["band_starts"][i], config["data"]["band_ends"][i], config["condor"]["data_load_size"] - 0.1)[:-1]
        bandend = bandstart + config["condor"]["data_load_size"]
        bandwidths =  [config["data"]["band_widths"][i]] * len(bandstart)
        tlist = np.array([bandstart, bandend, bandwidths]).T
        band_list.append(tlist)

    print(config["data"]["run"])

    sub_comment = f'{config["data"]["run"]}_{datatype}_split'
    
    sub_filename = "{}/{}.sub".format(condor_dir,sub_comment)
    write_subfile(sub_filename,config.config_file,config,datatype,sub_comment,dirs,verbose=False)
    dag_filename = "{}/{}.dag".format(condor_dir,sub_comment)

    with open(dag_filename,'w') as f:
        seeds = []
        for l_band in band_list:
            for sband in l_band:
                seeds.append(random.randint(1,1e9))
        for i,l_band in enumerate(band_list):
            for j,sband in enumerate(l_band):
                comment = "F_{}".format(sband[0])
                uid = seeds[i*j]
                jobid = "{}_{}_{}".format(comment,i*j,uid)
                job_string = "JOB {} {}\n".format(jobid,sub_filename)
                retry_string = "RETRY {} 10\n".format(jobid)
                vars_string = f'VARS {jobid} bandstart="{sband[0]}" bandend="{sband[1]}" bandwidth="{sband[2]}"\n'
                f.write(job_string)
                f.write(retry_string)
                f.write(vars_string)

    
def main():
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v', '--verbose', help='display status', action='store_true')
    parser.add_argument("-c", "--config-file", help="config file contatining parameters")
    #parser.add_argument("-d", "--data-type", help="datatype (train, val, test)", default="train")

    try:                                                     
        args = parser.parse_args()  
    except:  
        sys.exit(1)

    from soapcw.soap_config_parser import SOAPConfig

    config = SOAPConfig(args.config_file)

    for k, typename in enumerate(config["data"]["type"]):
        write_dagfile(config, typename)

if __name__ == '__main__':
    main()


    



from __future__ import print_function
import numpy as np
import matplotlib
try:
    import matplotlib.pyplot as plt
except:
    print("cannot import matplotlib")
import timeit
import scipy.stats as st
import os
import matplotlib.ticker as ticker

def plot_single(spect,soapout=None,fmin=None,fmax=None,tsft=None,tstart=None):
    """
    plot a spectrogtam with soap track overlaid
    """
    if fmin is None and fmax is None:
        fmin = 0
        fmax = len(spect[0])

    if tsft is None:
        tsft = len(spect[0])/(fmax-fmin)

    nsft = len(spect)
    tick_fontsize = 13
    label_fontsize = 20

    fig, ax = plt.subplots(nrows=2,figsize=(18,11),sharex=True)
    xticks = np.linspace(0,(nsft-1),10)
    xticklabels = np.linspace(0,(nsft-1)*tsft,10).astype(int)
    yticks = np.linspace(0,len(spect[0]),8).astype(int)
    yticklabels = np.round(np.linspace(fmin,fmax,8),2)
    plot_spect = np.ma.masked_where(spect == 96., spect,copy = True)
    
    im = ax[0].imshow(plot_spect.T,cmap="YlGnBu",aspect="auto",origin="lower")
    if soapout is not None:
        imv = ax[1].imshow(soapout.vitmap.T,cmap="YlGnBu",aspect="auto",origin="lower")
        cbarv = fig.colorbar(imv,ax = ax[1])
        cbarv.set_label("log-viterbi 'probability'",fontsize = label_fontsize)
        cbarv.ax.tick_params(labelsize=tick_fontsize)
        
    cbar = fig.colorbar(im,ax = ax[0])
    cbar.set_label("Normalised SFT power",fontsize = label_fontsize)
    cbar.ax.tick_params(labelsize=tick_fontsize)
    if soapout is not None:
        line = ax[0].scatter(np.arange(nsft),soapout.vit_track,color="red",label="Viterbi track",s=3)
    ax[0].set_xticks(xticks)
    ax[1].set_xticks(xticks)
    ax[0].set_yticks(yticks)
    ax[1].set_yticks(yticks)
    ax[0].set_xticklabels(xticklabels)
    ax[0].set_yticklabels(yticklabels)
    ax[1].set_xticklabels(xticklabels)
    ax[1].set_yticklabels(yticklabels)
    
    ax[1].set_xlabel("Time - {}[s]".format(tstart),fontsize=label_fontsize)
    ax[0].set_ylabel("Frequency [Hz]",fontsize=label_fontsize)
    ax[1].set_ylabel("Frequency [Hz]",fontsize=label_fontsize)
    ax[1].xaxis.set_tick_params(rotation=0,labelsize=tick_fontsize)
    ax[0].yaxis.set_tick_params(rotation=0,labelsize=tick_fontsize)
    ax[1].yaxis.set_tick_params(rotation=0,labelsize=tick_fontsize)
    #ax[1].xaxis.set_major_formatter(ticker.StrMethodFormatter("{x:.2f}"))
    ax[0].set_xlim([0,nsft])
    ax[1].set_xlim([0,nsft])
    ax[0].set_ylim([0,len(spect[0])])
    ax[1].set_ylim([0,len(spect[0])])

    for spine in list(ax[0].spines.values()) + list(ax[1].spines.values()):
        spine.set_visible(False)
        
    cbar.outline.set_visible(False)
    if soapout is not None:
        cbarv.outline.set_visible(False)
    fig.subplots_adjust(wspace=0, hspace=0)
    if soapout is not None:
        return fig, (im,imv,line)
    else:
        return fig, (im)



from soapcw.cnn import cnn_data_gen
import numpy as np
from soapcw import cw
import time
import os
import datetime
import copy
import pickle

def loop_band_test_data(loadpath,path,bandmin,bandmax,band_width,resize_image=True,search=False, num_repeat = 1):

    if 40 <= bandmin <= 500: 
        stride = 1
        degfree = 96
        plotcut = stride*180
        width = 0.1
        brange = "band_40_500"
    elif 500 < bandmin <= 1000:
        stride = 2
        degfree = int(stride*96)
        plotcut = stride*180
        width = 0.2
        brange = "band_500_1000"
    elif 1000 < bandmin <= 1500:
        stride = 3
        degfree = int(stride*96)
        plotcut = stride*180
        width = 0.3
        brange = "band_1000_1500"
    elif 1500 < bandmin <= 2000:
        stride = 4
        degfree = int(stride*96)
        plotcut = stride*180
        width = 0.4
        brange = "band_1500_2000"

        
    bl,be = bandmin,bandmax

    tmin, tmax = None,None#931035615.,931035615.+1800*100

    vetolist = list(np.arange(bandmin,bandmax))
    
    hname,lname = cnn_data_gen.find_sft_file(loadpath,bandmin,bandmax)

    sfts = cw.LoadSFT("{};{}".format(hname,lname),norm=True,filled=True,vetolist = vetolist,save_rngmed=True)

    bin_width = 1./float(1800.)
    nbins = len(getattr(sfts,"H1").frequencies)
    sub_bins = np.round(width/bin_width).astype(int) 

    # set as list for fill 2 Hz band
    start_time = time.time()
    range_bands = np.round(np.arange(bandmin,bandmax-width,0.5*width),2)
    #range_index = np.arange(0,(bandmax-width-bandmin)*1800,90).astype(int)
    range_index = np.round(np.arange(0,nbins,width/(2*bin_width))).astype(int)[:-1]

    # set which bands are odd and even with and ondex of 1 or 0
    odd_even = np.zeros(len(range_bands))
    odd_even[1::2] = 1

    if not os.path.isdir(path):
        os.makedirs(path)
    erange = 15

    snrstart,snrend = 20,200

    noise_out_snr = "snr_0.0_0.0"
    signal_out_snr = "snr_{}_{}".format(snrstart,snrend)
    #save_options = ["stats","paths", "pars", "powers"]
    save_options = ["stats","paths", "pars", "powers","vit_imgs","H_imgs","L_imgs"]
    # define all save directories for images, stats, pars, plots etc. out_snr only exists for training data

    even_noise_filenames = {}
    even_signal_filenames = {}
    odd_noise_filenames = {}
    odd_signal_filenames = {}

    even_noise_save_outs = {}
    even_signal_save_outs = {}
    odd_noise_save_outs = {}
    odd_signal_save_outs = {}

    for dirname in save_options:

        even_noise_save_outs[dirname] = []
        even_signal_save_outs[dirname] = []
        odd_noise_save_outs[dirname] = []
        odd_signal_save_outs[dirname] = []



    for i in range(len(range_bands)):
            
        k = range_bands[i]

        skip_band = False
        hardware_list = [(108.8,108.9),(147.45,147.55),(192.7,192.75)]
        for lin in hardware_list:
            if k > lin[0] and k < lin[1]:
                skip_band = True

        if skip_band:
            continue

        ms = int(range_index[i])
        me = int(ms+sub_bins)
        
        if odd_even[i] == 1:
            split = "odd"
        else:
            split = "even"

        #outpath = os.path.join(path,split)                     
            
        av_sh = (2/1800.)*np.nanmedian([sfts.H1.rng_med[:,ms:me],sfts.L1.rng_med[:,ms:me]])
        
        datah = copy.deepcopy(sfts.H1.norm_sft_power[:,ms:me])
        datal = copy.deepcopy(sfts.L1.norm_sft_power[:,ms:me])
        print("SFTLEN. H1:{}, L1:{}".format(sum(datah[:,0] != degfree),sum(datal[:,0] != degfree)))
        
        #print(k,k+width, snr )
        for nr in range(num_repeat):

            snr = int(np.random.rand(1)*(snrend-snrstart) + snrstart)
            print("SNR: ", snr)
            rnd = np.random.rand(1)

            if rnd > 0.5 or search:
                # if snr of 0 give downsamp data otherwise give sft for injections
                snr = 0
            
            outputs = run_and_inj_test(copy.copy(datah),copy.copy(datal),k,width,resize_image=resize_image,snr=snr,av_sh = av_sh,sfts=sfts, stride=stride, degfree=degfree, brange = brange)

            for key in save_options:
                if split == "even":
                    if snr == 0:
                        even_noise_save_outs[key].append(outputs[key])
                    else:
                        even_signal_save_outs[key].append(outputs[key])
                elif split == "odd":
                    if snr == 0:
                        odd_noise_save_outs[key].append(outputs[key])
                    else:
                        odd_signal_save_outs[key].append(outputs[key])

            
        del datah,datal


    for key in save_options:
        even_noise_filenames[key] = os.path.join(*[path,"even",key,brange,noise_out_snr,"freq_{}_{}.pkl".format(bandmin, len(even_noise_save_outs[key]))])
        even_signal_filenames[key] = os.path.join(*[path,"even",key,brange,signal_out_snr,"freq_{}_{}.pkl".format(bandmin, len(even_signal_save_outs[key]))])
        odd_noise_filenames[key] = os.path.join(*[path,"odd",key,brange,noise_out_snr,"freq_{}_{}.pkl".format(bandmin, len(odd_noise_save_outs[key]))])
        odd_signal_filenames[key] = os.path.join(*[path,"odd",key,brange,signal_out_snr,"freq_{}_{}.pkl".format(bandmin, len(odd_signal_save_outs[key]))])
    
        for fname in [even_noise_filenames[key], even_signal_filenames[key], odd_noise_filenames[key], odd_signal_filenames[key]]:
            if not os.path.isdir(os.path.dirname(fname)):
                os.makedirs(os.path.dirname(fname))

        print(even_noise_filenames[key])
        with open(even_noise_filenames[key],"wb") as f:
            pickle.dump(even_noise_save_outs[key], f)
        with open(odd_noise_filenames[key],"wb") as f:
            pickle.dump(odd_noise_save_outs[key], f)
        with open(even_signal_filenames[key],"wb") as f:
            pickle.dump(even_signal_save_outs[key], f)
        with open(odd_signal_filenames[key],"wb") as f:
            pickle.dump(odd_signal_save_outs[key], f)



    end_time = time.time()
    date = datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    with open(os.path.join(path,"timing.txt"),"a") as f:
        f.write("time-for-gen    {}    {} \n".format(end_time-start_time,date))

        
        
def run_and_inj_test(datah,datal,fmin,width,resize_image=False,snr = 0, av_sh = None,sfts=None, stride = 1, degfree = 96, brange = ""):
    
    
    # save noise realisation 1
    if snr == 0:
        nsft, tstart, tsft, flow, fhigh = len(sfts.H1.epochs), sfts.H1.epochs[0], 1800., fmin,fmin+width
        sig = cw.GenerateSignal()
        sig.alpha = 0
        sig.delta = 0
        sig.cosi = 0
        sig.phi0 = 0
        sig.psi = 0
        sig.f = [100,-1e-16]
        sig.snr = 0
        sig.fmin = flow
        sig.fmax = flow + width
        sig.tref = tstart

        sig.earth_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/earth00-40-DE430.dat.gz"
        sig.sun_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/sun00-40-DE430.dat.gz"
        data = sig.get_spectrogram(epochs = sfts.H1.epochs, nsft = len(datah),tref=sfts.H1.epochs[0],tsft=tsft,fmin=flow,fmax=fhigh,snr=0,noise_spect={"H1":datah,"L1":datal})
        #print(data.h0)
        data.sum_sfts()
        data.downsamp_frequency(stride=stride,data_type="summed_norm_sft_power",remove_original=False)
        H1path,L1path = (flow + width/2.,flow + width/2.),(flow + width/2.,flow + width/2.)
        save_pars = {"fmin":sig.fmin,"fmax":sig.fmax, "alpha":None, "delta":None, "cosi":None, "psi":None, "phi0":None, "f0":None, "fdot":None, "snr":0, "depth":np.inf, "tref":sig.tref, "av_sh": av_sh, "h0" : 0}

        noise_outputs = cnn_data_gen.return_outputs(out_snr="./",datah=data.H1.downsamp_summed_norm_sft_power,datal=data.L1.downsamp_summed_norm_sft_power,fmin=fmin,fmax=fmin+width,snr=0,depth=0,h0=0,resize_image=resize_image,test=True,epochs=data.H1.summed_epochs,minmaxpaths=(H1path,L1path),degfree=degfree, brange = brange, pars = save_pars)
        return noise_outputs
    else:
        # inject signal into band
        
        sig = cw.GenerateSignal()
        nsft, tstart, tsft, flow, fhigh = len(sfts.H1.epochs), sfts.H1.epochs[0], 1800., fmin,fmin+width
        
        #sigfreq = np.random.uniform(0,1)*(width)*0.5 + flow + (width)*0.2
        
        
        # freq, spindown^n, alpha, sin(delta), phi0, psi, cosi
        params_low =  [fmin + 0.03,      -9 ,0      ,-1,0      ,0      ,-1.0]
        params_high = [fmin + width-0.03,-16,2*np.pi,1 ,2*np.pi,np.pi/2,1.0]
        data_list_signal = []
        data_list_noise = []
        nsft = int(nsft)
        param_list = ["f","fd","Alpha","Delta","phi0","psi","cosi"]
        pars = {}
        for j in np.linspace(0,len(param_list)-1,len(param_list)).astype(int):
            pars[param_list[j]] = params_low[j] + np.random.uniform(0,1)*(params_high[j] - params_low[j])
            
        sig.alpha = pars["Alpha"]
        sig.delta = np.arcsin(pars["Delta"])
        sig.cosi = pars["cosi"]
        sig.phi0 = pars["phi0"]
        sig.psi = pars["psi"]
        sig.f = [pars["f"],-np.random.uniform(0,1)*(10**pars["fd"])]
        sig.earth_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/earth00-40-DE430.dat.gz"
        sig.sun_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/sun00-40-DE430.dat.gz"
        sig.snr = snr
        sig.fmin = flow
        sig.fmax = fhigh
        sig.tref = tstart



        #print(tstart, len(datah), tstart, tsft, flow,fhigh,snr,np.mean(datah),np.mean(datal))
        #print("FREQS", flow, fhigh)
        data = sig.get_spectrogram(epochs = sfts.H1.epochs, nsft = len(datah),tref=sfts.H1.epochs[0],tsft=tsft,fmin=flow,fmax=fhigh,snr=snr,noise_spect={"H1":datah,"L1":datal})
        #print(data.h0)
        data.sum_sfts()
        data.downsamp_frequency(stride=stride,data_type="summed_norm_sft_power",remove_original=False)

        H1pulsar_path = sig.get_pulsar_path(sfts.H1.epochs,sig.edat,"H1")
        L1pulsar_path = sig.get_pulsar_path(sfts.L1.epochs,sig.edat,"L1")

        H1paths = min(H1pulsar_path),max(H1pulsar_path)
        L1paths = min(L1pulsar_path),max(L1pulsar_path)

        depth = data.depth
        h0 = np.sqrt(av_sh)/depth
        save_pars = {"fmin":sig.fmin,"fmax":sig.fmax, "alpha":sig.alpha, "delta":sig.delta, "cosi":sig.cosi, "psi":sig.psi, "phi0":sig.phi0, "f0":sig.f[0], "fdot":sig.f[1], "snr":snr, "depth":depth, "tref":sig.tref, "av_sh": av_sh, "h0" : h0}
        

        sig_outputs = cnn_data_gen.return_outputs(out_snr="./",datah=data.H1.downsamp_summed_norm_sft_power,datal=data.L1.downsamp_summed_norm_sft_power,fmin=fmin,fmax=fmin+width,snr=snr, h0=h0,depth=depth,resize_image=resize_image,pars=save_pars,test=True,epochs=data.H1.summed_epochs,minmaxpaths=(H1paths,L1paths),degfree=degfree, brange = brange)
        
        return sig_outputs
        



import torch
import torch.nn as nn
import numpy as np

class CNN(nn.Module):

    def __init__(self, input_dim, fc_layers = [], conv_layers = [], stride = 1, device="cpu", dropout=0.1, inchannels = 1, avg_pool_size=None):
        """
        args
        ----------
        input_dim: int
            length of input time series
        latent_dim: int
            number of variables in latent space
        fc_layers: list
            list of the fully connected layers [8, 4, 2] (8 neurons, 4 neurons, 2 neurons)
        conv_layers: list
            list of the convolutional layers [(8, 5, 2, 1), (4, 3, 2, 1)] (8 filters of size 5, 4 filters of size 3)
        """
        super().__init__()
        # define useful variables of network
        self.device = device
        self.input_dim = input_dim
        self.inchannels = inchannels
        
        # convolutional parts
        self.fc_layers = fc_layers
        self.conv_layers = conv_layers
        self.num_conv = len(self.conv_layers)
        self.stride = stride
        self.avg_pool_size = tuple(avg_pool_size)

        self.activation = nn.LeakyReLU()
        self.out_activation = nn.Sigmoid()#nn.Softmax()
        self.drop = nn.Dropout(p=dropout)

        self.small_const = 1e-6
        
        self.conv_network, self.lin_network = self.create_network("o", self.input_dim, fc_layers = self.fc_layers, conv_layers = self.conv_layers)


    def create_network(self, name, input_dim,  append_dim=0, fc_layers=[], conv_layers=[]):
        """ Generate arbritrary network, with convolutional layers or not
        args
        ------
        name: str
            name of network
        input_dim: int
            size of input to network
        append_dim: int (optional)
            number of neurons to append to fully connected layers after convolutional layers
        mean: bool (optional) 
            if True adds extra output layer for means 
        variance : bool (optional)
            if True adds extra layer of outputs for variances
        fc_layers: list
            list of fully connected layers, format: [64,64,...] = [num_neurons_layer1, num_neurons_layer2, .....]
        conv_layers: list
            list of convolutional layers, format:[(8,8,2,1), (8,8,2,1)] = [(num_filt1, conv_size1, max_pool_size1, dilation1), (num_filt2, conv_size2, max_pool_size2, dilation2)] 
        """
        conv_network = nn.Sequential() # initialise networks
        lin_network = nn.Sequential()
        layer_out_sizes = []
        num_conv = len(conv_layers)
        num_fc = len(fc_layers)
        insize = self.input_dim
        inchannels = self.inchannels
        # add convolutional layers
        for i in range(num_conv):
            # set layer parameters
            padding = (int(conv_layers[i][1]/2.),int(conv_layers[i][1]/2.)) # padding half width
            maxpool_size = (conv_layers[i][2],conv_layers[i][2]) #define max pooling for this layer
            filt_size = (conv_layers[i][1],conv_layers[i][1])
            stride = (self.stride, self.stride)
            dilation = (conv_layers[i][3],conv_layers[i][3])
            maxpool = nn.MaxPool2d(maxpool_size) #define max pooling for this layer
            # add convolutional/activation/maxpooling layers
            #print(inchannels, conv_layers[i][0], filt_size, stride, padding, dilation)
            conv_network.add_module("r_conv{}".format(i),module = nn.Conv2d(inchannels, conv_layers[i][0], filt_size, stride = stride, padding=padding, dilation = dilation))
            conv_network.add_module("act_r_conv{}".format(i), module = self.activation)
            conv_network.add_module("pool_r_conv{}".format(i),module = maxpool)
            # define the output size of the layer
            outsize = self.conv_out_size(insize[0], insize[1], padding, dilation, filt_size, stride, maxpool_size) # output of one filter
            layer_out_sizes.append((conv_layers[i][0],outsize[0],outsize[1]))
            insize = outsize
            inchannels = conv_layers[i][0]

        if self.avg_pool_size is not None and num_conv > 0:
            conv_network.add_module("avgpool", nn.AdaptiveAvgPool2d(self.avg_pool_size))
            lin_input_size = np.prod(self.avg_pool_size)*layer_out_sizes[-1][0]
        else:
            # define the input size to fully connected layer
            lin_input_size = np.prod(layer_out_sizes[-1]) if num_conv > 0 else np.prod(self.input_dim)

        if append_dim:
            lin_input_size += append_dim
        layer_size = int(lin_input_size)
        # hidden layers
        for i in range(num_fc):
            lin_network.add_module("r_lin{}".format(i),module=nn.Linear(layer_size, fc_layers[i]))
            if i == num_fc - 1:
                pass
                #lin_network.add_module("act_r_lin{}".format(i),module=self.out_activation)
            else:
                lin_network.add_module("r_drop{}".format(i),module=self.drop)
                lin_network.add_module("act_r_lin{}".format(i),module=self.activation)
            layer_size = fc_layers[i]

        return conv_network, lin_network


    def conv_out_size(self, height_in, width_in, padding, dilation, kernel, stride, maxpool_size):
        height_out = int(((height_in + 2*padding[0] - dilation[0]*(kernel[0]-1)-1)/stride[0] + 1)/maxpool_size[0])
        width_out = int(((width_in + 2*padding[1] - dilation[1]*(kernel[1]-1)-1)/stride[1] + 1)/maxpool_size[1])
        return height_out, width_out
        
    def forward(self, y):
        """forward pass for training"""
        #y = y, (-1, self.inchannels, self.input_dim[0], self.input_dim[1])) if self.num_conv > 0 else y
        conv = self.conv_network(y) if self.num_conv > 0 else y
        lin_in = torch.flatten(conv,start_dim=1) if self.num_conv > 0 else y   # flatten convolutional layers
        out = self.lin_network(lin_in) # run fully connected network
        return out
    
    def test(model, y):
        """generating samples when testing the network """
        num_data = y.size(0)                                                                                                                                                     
        x_out = []
        # encode the data into latent space with r1(z,y)          
        for i in range(num_data):
            x_out.append(model.forward(y[i]).cpu().numpy())
        return np.array(x_out)






class VitmapSpectStatCNN(nn.Module):

    def __init__(self, input_dim, output_dim, fc_layers = [], conv_layers = [], stride = 1, device="cpu", dropout=0.0, inchannels = 1, stat_network = None, vitmap_network = None, spect_network = None, train_loaded_model = False):
        """
        args
        ----------
        input_dim: int
            length of input time series
        latent_dim: int
            number of variables in latent space
        output_dim: int
            number of output neurons
        fc_layers: list
            list of the fully connected layers [8, 4, 2] (8 neurons, 4 neurons, 2 neurons)
        conv_layers: list
            list of the convolutional layers [(8, 5, 2, 1), (4, 3, 2, 1)] (8 filters of size 5, 4 filters of size 3)
        """
        super().__init__()
        # define useful variables of network
        self.device = device
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.inchannels = inchannels

        # Define networks to load in
        self.stat_network = stat_network
        self.vitmap_network = vitmap_network
        self.spect_network = spect_network
        
        # convolutional parts
        self.fc_layers = fc_layers
        self.conv_layers = conv_layers
        self.num_conv = len(self.conv_layers)
        self.stride = stride

        self.activation = nn.LeakyReLU()
        self.out_activation = nn.Sigmoid()#nn.Softmax()
        self.drop = nn.Dropout(p=dropout)

        self.small_const = 1e-6

        # load or generate the spectrogram network
        if self.spect_network is not None:
            self.spect_network = torch.load(self.spect_network, map_location=self.device)
        else:
            self.spect_network = CNN(self.input_dim, self.output_dim, fc_layers = self.fc_layers, conv_layers =self.conv_layers, stride = self.stride, dropout = self.dropout, inchannels = 2)

        #load or generate the vitmap network
        if self.vitmap_network is not None:
            self.vitmap_network = torch.load(self.vitmap_network, map_location=self.device)
        else:
            self.vitmap_network = CNN(self.input_dim, self.output_dim, fc_layers = self.fc_layers, conv_layers =self.conv_layers, stride = self.stride, dropout = self.dropout, inchannels = 1)

        # load or generate the line aware stat network
        if self.stat_network is not None:
            self.stat_network = torch.load(self.stat_network, map_location = self.device)
        else:
            self.stat_network = nn.Sequential()
            self.stat_network.add_module("lin1",module=nn.Linear(1, 1))
            self.stat_network.add_module("act_lin1",module=self.out_activation)

        self.spect_network.lin_network = nn.Sequential(*list(self.spect_network.lin_network.children())[:-1])
        self.vitmap_network.lin_network = nn.Sequential(*list(self.vitmap_network.lin_network.children())[:-1])
        
        for net in [self.spect_network.lin_network, self.spect_network.conv_network, self.vitmap_network.lin_network, self.vitmap_network.conv_network]:
            for param in list(net.parameters()):
                param.requires_grad = train_loaded_model

        self.final_classifier = nn.Sequential()
        # we remove last layer from spect and vitmap networks so their outputs are of size fc_layers[-1]
        #we then add this to the output of stat network
        self.final_classifier.add_module("linout",module=nn.Linear(2*self.fc_layers[-1] + 1, self.output_dim))
        self.final_classifier.add_module("act_linout",module=self.out_activation)

        
    def forward(self, y_vitmap, y_spect, y_stat):
        """forward pass for training"""
        # run the vitmap network
        y_vitmap = torch.reshape(y_vitmap, (-1, self.vitmap_network.inchannels, self.input_dim[0], self.input_dim[1])) 
        vitmap_conv = self.vitmap_network.conv_network(y_vitmap)
        vitmap_lin = torch.flatten(vitmap_conv,start_dim=1)
        vitmap_out = self.vitmap_network.lin_network(vitmap_lin) # run fully connected network

        # run the spect network
        y_spect = torch.reshape(y_spect, (-1, self.spect_network.inchannels, self.input_dim[0], self.input_dim[1])) 
        spect_conv = self.spect_network.conv_network(y_spect)
        spect_lin = torch.flatten(spect_conv,start_dim=1)
        spect_out = self.spect_network.lin_network(spect_lin) # run fully connected network

        # run the stat network
        y_stat = torch.reshape(y_stat, (-1, 1)) 
        stat_out = self.stat_network(y_stat) # run fully connected network
        
        # combine all three outputs
        out = self.final_classifier(torch.cat([vitmap_out, spect_out, stat_out], dim = 1))

        return out
    
    def test(model, y_vitmap, y_spect, y_stat):
        """generating samples when testing the network """
        num_data = y_vitmap.size(0)                                                                                                                                                     
        x_out = []
        # encode the data into latent space with r1(z,y)          
        for i in range(num_data):
            x_out.append(model.forward(y_vitmap[i], y_spect[i], y_stat[i]).cpu().numpy())
        return np.array(x_out)




class VitmapStatCNN(nn.Module):

    def __init__(self, input_dim, output_dim, fc_layers = [], conv_layers = [], stride = 1, device="cpu", dropout=0.0, inchannels = 1, stat_network = None, vitmap_network = None, train_loaded_model = False):
        """
        args
        ----------
        input_dim: int
            length of input time series
        latent_dim: int
            number of variables in latent space
        output_dim: int
            number of output neurons
        fc_layers: list
            list of the fully connected layers [8, 4, 2] (8 neurons, 4 neurons, 2 neurons)
        conv_layers: list
            list of the convolutional layers [(8, 5, 2, 1), (4, 3, 2, 1)] (8 filters of size 5, 4 filters of size 3)
        """
        super().__init__()
        # define useful variables of network
        self.device = device
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.inchannels = inchannels

        # Define networks to load in
        self.stat_network = stat_network
        self.vitmap_network = vitmap_network
        
        # convolutional parts
        self.fc_layers = fc_layers
        self.conv_layers = conv_layers
        self.num_conv = len(self.conv_layers)
        self.stride = stride

        self.activation = nn.LeakyReLU()
        self.out_activation = nn.Sigmoid()#nn.Softmax()
        self.drop = nn.Dropout(p=dropout)

        self.small_const = 1e-6

        #load or generate the vitmap network
        if self.vitmap_network is not None:
            self.vitmap_network = torch.load(self.vitmap_network, map_location=self.device)
        else:
            self.vitmap_network = CNN(self.input_dim, self.output_dim, fc_layers = self.fc_layers, conv_layers =self.conv_layers, stride = self.stride, dropout = self.dropout, inchannels = 1)

        # load or generate the line aware stat network
        if self.stat_network is not None:
            self.stat_network = torch.load(self.stat_network, map_location = self.device)
        else:
            self.stat_network = nn.Sequential()
            self.stat_network.add_module("lin1",module=nn.Linear(1, 1))
            self.stat_network.add_module("act_lin1",module=self.out_activation)

        # remove last layer of vitmap network
        self.vitmap_network.lin_network = nn.Sequential(*list(self.vitmap_network.lin_network.children())[:-1])
        
        for net in [self.vitmap_network.lin_network, self.vitmap_network.conv_network]:
            for param in list(net.parameters()):
                param.requires_grad = train_loaded_model

        self.final_classifier = nn.Sequential()
        # we remove last layer from vitmap networks so their outputs are of size fc_layers[-1]
        #we then add this to the output of stat network
        self.final_classifier.add_module("linout",module=nn.Linear(self.fc_layers[-1] + 1, self.output_dim))
        self.final_classifier.add_module("act_linout",module=self.out_activation)

        
    def forward(self, y_vitmap, y_stat):
        """forward pass for training"""
        # run the vitmap network
        y_vitmap = torch.reshape(y_vitmap, (-1, self.vitmap_network.inchannels, self.input_dim[0], self.input_dim[1])) 
        vitmap_conv = self.vitmap_network.conv_network(y_vitmap)
        vitmap_lin = torch.flatten(vitmap_conv,start_dim=1)
        vitmap_out = self.vitmap_network.lin_network(vitmap_lin) # run fully connected network

        # run the stat network
        y_stat = torch.reshape(y_stat, (-1, 1)) 
        stat_out = self.stat_network(y_stat) # run fully connected network
        
        # combine all three outputs
        out = self.final_classifier(torch.cat([vitmap_out,  stat_out], dim = 1))

        return out
    
    def test(model, y_vitmap, y_stat):
        """generating samples when testing the network """
        num_data = y_vitmap.size(0)                                                                                                                                                     
        x_out = []
        # encode the data into latent space with r1(z,y)          
        for i in range(num_data):
            x_out.append(model.forward(y_vitmap[i], y_stat[i]).cpu().numpy())
        return np.array(x_out)




#!/usr/bin/python
import sys
import os
import getopt
import time
import configparser 
import argparse
import random
import numpy as np
from astropy.time import Time
import json
from .soap_config_parser import SOAPConfig


def write_plot_subfile(sub_filename,config,config_file,cdirs,comment,verbose=True):

    with open(sub_filename,'w') as f:
        f.write('# filename: {}\n'.format(sub_filename))
        f.write('universe = vanilla\n')
        #f.write('executable = {}\n'.format(config["scripts"]["search_exec"]))
        # finds the location of the python executable and replaces with the correct soap exec
        execute = os.path.join(os.path.split(sys.executable)[0],config["scripts"]["search_exec"])
        f.write('executable = {}\n'.format(execute))
        f.write('getenv  = True\n')
        f.write('RequestMemory = {} \n'.format(config["condor"]["memory"]))
        f.write(f'request_disk={config["condor"]["request_disk"]}\n')
        f.write('log = {}/{}_$(cluster).log\n'.format(cdirs["log_dir"],comment))
        f.write('error = {}/{}_$(cluster).err\n'.format(cdirs["err_dir"],comment))
        f.write('output = {}/{}_$(cluster).out\n'.format(cdirs["output_dir"],comment))
        args = f'arguments = --config-file {config_file} -s $(bandstart) -e $(bandend) -r {config["data"]["obs_run"]} -w {config["condor"]["band_load_size"]} -sd {config["output"]["sub_directory"]}'

        f.write(args + "\n")
        #f.write('accounting_group = aluk.dev.s6.cw.viterbi\n')
        f.write(f'accounting_group = {config["condor"]["accounting_group"]}\n')
        f.write('queue\n')
    if verbose:
        print(time.time(), "generated subfile")
                
def write_html_subfile(sub_filename,config, config_file, cdirs, comment, verbose=True):

    with open(sub_filename,'w') as f:
        f.write('# filename: {}\n'.format(sub_filename))
        f.write('universe = vanilla\n')
        f.write('executable = {}\n'.format(config["scripts"]["html_exec"]))
        #f.write('enviroment = ""\n')
        f.write('getenv  = True\n')
        f.write('RequestMemory = {} \n'.format(config["condor"]["memory"]))
        f.write(f'request_disk={config["condor"]["request_disk"]}\n')
        f.write('log = {}/{}_$(cluster).log\n'.format(cdirs["log_dir"],comment))
        f.write('error = {}/{}_$(cluster).err\n'.format(cdirs["err_dir"],comment))
        f.write('output = {}/{}_$(cluster).out\n'.format(cdirs["output_dir"],comment))
        
        args = f'arguments = -c {config_file}'
        f.write(args + "\n")

        #f.write('accounting_group = aluk.dev.s6.cw.viterbi\n')
        f.write(f'accounting_group = {config["condor"]["accounting_group"]}\n')
        f.write('queue\n')
    if verbose:
        print(time.time(), "generated subfile")


def write_dagfile(config_file, config, verbose = False):
    """ Write a dag file containing all link to submit file for all sub bands"""

    condor_dirs = {}
    condor_dirs["condor_dir"] = os.path.join(config["general"]["root_dir"], "condor")

    condor_dirs["log_dir"] = os.path.join(condor_dirs["condor_dir"],"logs")
    condor_dirs["err_dir"] = os.path.join(condor_dirs["condor_dir"],"err")
    condor_dirs["output_dir"] = os.path.join(condor_dirs["condor_dir"],"output")
    
    condor_dirs["save_directory"] = os.path.join(config["output"]["save_directory"], config["data"]["obs_run"], config["output"]["sub_directory"])
    
    for key, val in condor_dirs.items():
        if not os.path.isdir(val):
            os.makedirs(val)

    ##
    ## Make the sub file for SOAP run and summary page generation
    ##

    bs1 = config["data"]["band_starts"]
    be1 = config["data"]["band_ends"]

    run_sub_comment = "soap_line_F{}_{}_{}".format(bs1[0],be1[0],config["data"]["obs_run"])
    html_sub_comment = "soap_html_F{}_{}_{}".format(bs1[0],be1[0],config["data"]["obs_run"])

    run_sub_filename = "{}/{}.sub".format(condor_dirs["condor_dir"],run_sub_comment)
    html_sub_filename = "{}/{}.sub".format(condor_dirs["condor_dir"],html_sub_comment)

    write_plot_subfile(run_sub_filename,config,config_file,condor_dirs,run_sub_comment,verbose=verbose)
    write_html_subfile(html_sub_filename,config,config_file,condor_dirs,html_sub_comment,verbose=verbose)

    ##
    ## Make list of band starts from condor split
    ##
    band_starts = []
    band_ends = []
    for i, bs in enumerate(bs1):
        t_bss = np.arange(bs1[i],
            be1[i],
            config["condor"]["band_load_size"]
            )[:-1]
        t_bes = t_bss + config["condor"]["band_load_size"] + config["data"]["band_widths"][i]

        band_starts.extend(list(t_bss))
        band_ends.extend(list(t_bes))

    dag_filename = "{}/{}.dag".format(condor_dirs["condor_dir"],run_sub_comment)

    jobids = []
    with open(dag_filename,'w') as f:
        seeds = []
        for i in range(len(band_starts)):
            seeds.append(random.randint(1,1e9))

        for i in range(len(band_starts)):
            comment = "F_{}".format(int(band_starts[i]))
            jobid = "{}_{}_{}".format(comment,i,seeds[i])
            jobids.append(jobid)
            job_string = "JOB {} {}\n".format(jobid,run_sub_filename)
            retry_string = "RETRY {} 2\n".format(jobid)
            vars_string = 'VARS {} bandstart="{}" bandend="{}"\n'.format(jobid,band_starts[i],band_ends[i])
            f.write(job_string)
            f.write(retry_string)
            f.write(vars_string)
            """
            if i == 0:
                order_string = None
            else:
                prev_job_ids = ""
                prev_uid = seeds[i-1]
                prev_comment = "F_{}".format(band_starts[i-1])
                prev_jobid = " {}_{}_{}".format(prev_comment,i-1,prev_uid)
                order_string = "PARENT {} CHILD {} \n".format(prev_jobid,jobid)
                #f.write(order_string)
            """
        
        comment = "F_{}".format("html")
        jobid = "{}_{}".format(comment,1)
        job_string = "JOB {} {}\n".format(jobid,html_sub_filename)
        retry_string = "RETRY {} 1\n".format(jobid)
        #vars_string = 'VARS \n'
        #prev_jobid = "{}".format([i jobids])
        order_string = "PARENT {} CHILD {} \n".format(" ".join(jobids),jobid)
        f.write(job_string)
        f.write(retry_string)
        #f.write(vars_string)
        f.write(order_string)
    if verbose:
        print(time.time(), "generated dag file")
    

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config-file", help="configuration file for soap")
    try:                                                     
        args = parser.parse_args()  
    except:  
        sys.exit(1)

    config_file = os.path.abspath(args.config_file)

    config = SOAPConfig(args.config_file)
    #config = configparser.ConfigParser()
    #config.read(config_file)

    write_dagfile(config_file, config)

if __name__ == '__main__()':
    main()


    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
from setuptools.command.install import install
import subprocess
from distutils.extension import Extension
import os


def build_ext(*args,**kwargs):
    from Cython.Distutils import build_ext as _build_ext
    return _build_ext(*args,**kwargs)
    
def cythonize(*args,**kwargs):
    from Cython.Build import cythonize as _cythonize
    return _cythonize(*args,**kwargs)


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()
    
class CustomInstall(install):
    def run(self):
        command = "git clone https://github.com/boostorg/boost.git"
        #process = subprocess.Popen(command, shell=True, cwd="soapcw/line_aware_statistic")
        #process.wait()

        command1 = "git clone https://gitlab.com/GNU/GSL.git"
        #process1 = subprocess.Popen(command1, shell=True, cwd="soapcw/line_aware_statistic")
        #process1.wait()

        install.run(self)
    
    
cwd = os.getcwd()
homedir = os.path.expanduser("~")


soapcw_include_dirs = ["{}/soapcw/glibc_fix.h".format(cwd),
                     "{}/repositories/gsl/include/".format(homedir),
                     "{}/repositories/boost_1_71_0/boost/math/distributions/".format(homedir)]

int_include_dirs = ["/{}/repositories/boost_1_71_0/boost/math/distributions/".format(homedir),
                    "/{}/repositories/boost_1_71_0/".format(homedir),
                    "/{}/repositories/".format(homedir),
                    "{}/soapcw/glibc_fix.h".format(cwd),
                    "{}/soapcw/line_aware_statistic/".format(homedir),
                    "/{}/repositories/gsl/include/".format(homedir)]

int_libraries = ["gsl","gslcblas"]

int_library_dirs = ["/{}/repositories/gsl/lib/".format(homedir)]




dep_links = []

# will include this later
lookup_module =     Extension("soapcw.line_aware_stat.gen_lookup",
                              ["soapcw/line_aware_stat/gen_lookup.pyx", "soapcw/line_aware_stat/integrals.cpp"],
                              language='c++',
                              include_dirs=int_include_dirs,
                              library_dirs=int_library_dirs,
                              libraries=int_libraries)

# decide whether to build lookup tables using the C++ code (is faster but requires external libraries)
cpp=False

if cpp:
    ext_modules = [
        Extension("soapcw.soapcw",
                  [ "soapcw/soapcw.pyx" ]),
        lookup_module,]
else:
    ext_modules = [
    Extension("soapcw.soapcw",
              [ "soapcw/soapcw.pyx" ]),]


cmdclass = { 'build_ext': build_ext , }

setup_requirements = ["cython", "matplotlib", "numpy", "torch", "torch-summary", "scipy", "lalsuite", "regex"]

test_requirements = ["cython"]

setup(
    author="Joe Bayley",
    author_email='joseph.bayley@glasgow.ac.uk',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="SOAP is a rapid algorithm to search for continuous sources of gravitational waves, with a wider application to long duration narrowband signals.",
    install_requires=setup_requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    #package_data = {"":['cnn/pytorch/trained_models/O3/*.ckpt','cnn/pytorch/trained_models/O3/*.ini']},
    keywords=['soapcw','soap','gravitational waves','pulsars','neutron stars','continuous waves'],
    name='soapcw',
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://git.ligo.org/joseph.bayley/soapcw',
    download_url='https://git.ligo.org/joseph.bayley/soapcw/-/archive/0.1.2/soapcw-0.1.2.tar.gz',
    zip_safe=False,
    ext_modules = cythonize(ext_modules),
    cmdclass=cmdclass,
)



from soapcw.cnn import cnn_data_gen
import numpy as np
from soapcw import cw
import time
import os
import datetime
import copy
import pickle
import h5py

def roll_in_time(indata,numbins = 100,sh=None):
    """Roll the data in time by a given number of bins, preserving the location of the gaps

    Args:
        indata (2d array): input data
        numbins (int, optional): number of time bins to roll by. Defaults to 100.
        sh (_type_, optional): the noise floor. Defaults to None.

    Returns:
        _type_: _description_
    """
    sfts_index = [i for i,j in enumerate(np.mean(indata,axis=1) != 2) if j]
    
    sfts_vals_rolled = np.roll(indata[sfts_index].T,numbins).T
    
    if sh is not None:
        sh_rolled = np.roll(sh[sfts_index].T,numbins).T
        sh_new = np.ones(len(sh))*np.nan
    
    data = np.ones(np.shape(indata))*2
    
    for i,j in enumerate(sfts_index):
        data[j,:] = sfts_vals_rolled[i]
        if sh is not None:
            sh_new[j] = sh_rolled[i]
        
    if sh is not None:
        return data, sh_new
    else:
        return data
        
def flip_data_time(indata,sh = None):
    """Flip the data in time preserving the location of the gaps

    Args:
        indata (2d array): input data
        sh (_type_, optional): noisefloor. Defaults to None.

    Returns:
        _type_: _description_
    """
    sfts_index = [i for i,j in enumerate(np.mean(indata,axis=1) != 2) if j]
    
    sfts_vals_reverse = indata[sfts_index][::-1]
    
    data = np.ones(np.shape(indata))*2
    
    if sh is not None:
        sh_reverse = sh[sfts_index][::-1]
        sh_new = np.ones(len(sh))*np.nan
    
    for i,j in enumerate(sfts_index):
        data[j,:] = sfts_vals_reverse[i]
        if sh is not None:
            sh_new[j] = sh_reverse[i]
        
    if sh is not None:
        return data, sh_new
    else:
        return data

def flip_data_freq(indata):
    """flip the data in frequency

    Args:
        indata (2d array): input spectrogram

    Returns:
        _type_: _description_
    """
    data = indata.T[::-1].T
        
    return data



def loop_band_train_augment(
    config,
    path,
    bandmin,
    bandmax,
    band_width,
    resize_image=True,
    gen_noise_only=False,
    snr_min = 50,
    snr_max=150, 
    save_options = None):
    """Loop over the data in load directory and augment to create training data

    Args:
        path (_type_): _description_
        bandmin (_type_): _description_
        bandmax (_type_): _description_
        band_width (_type_): _description_
        resize_image (bool, optional): _description_. Defaults to True.
        gen_noise_only (bool, optional): _description_. Defaults to False.
        snr_min (int, optional): _description_. Defaults to 50.
        snr_max (int, optional): _description_. Defaults to 150.
        save_options (_type_, optional): _description_. Defaults to None.
    """

    start_time = time.time()
    tsft = 1800.

    for i, (minband, maxband) in enumerate(zip(config["data"]["band_starts"], config["data"]["band_ends"])):
        if minband <= bandmin < maxband:
            stride = config["data"]["strides"][i]
            degfree = int(2 * config["data"]["n_summed_sfts"] * stride)
            width = config["data"]["band_widths"][i]
            nbin = 180#int(width*tsft)
            plotcut = stride * nbin
            brange = f"band_{int(minband)}_{int(maxband)}"
            break
        else:
            continue



    # in 2Hz band set frequencies and indicies with 0.1 Hz bands not overlapping
    range_bands = np.round(np.arange(bandmin,bandmax-width,width),1)
    range_index = np.arange(0,(bandmax-width-bandmin)*int(tsft),nbin*stride).astype(int)    
    
    # set which bands are odd and even with and ondex of 1 or 0
    odd_even = np.zeros(len(range_bands))
    odd_even[1::2] = 1

    bl,be = bandmin,bandmax

    if config["data"]["tstart"] is not None:
        tmin = config["data"]["tstart"]
    else:
        tmin = None
    if config["data"]["tend"] is not None:
        tmax = config["data"]["tend"]
    else:
        tmax = None

    #tmin, tmax = None,None#931035615.,931035615.+1800*100

    # set integer Hz as vetos, this will cut out a few bins surrounding
    vetolist = list(np.arange(bandmin,bandmax))

    # find the appropriate sft files
    hname,lname = cnn_data_gen.find_sft_file(config["general"]["narrowband_sft_dir"],bandmin,bandmax)
    
    # load in the narrowbanded sfts, normalise them, fill in the gaps and save the running median
    sfts = cw.LoadSFT("{};{}".format(hname,lname),norm=True,filled=True,vetolist = vetolist,save_rngmed=True, tmin=tmin,tmax=tmax)

    print("SFTs Loaded ...")

    # make appropriate directoes
    if not os.path.isdir(path):
        os.makedirs(path)
    erange = 15

    snrrange = [snr_min,snr_max]
    noise_out_snr = "snr_0.0_0.0"
    signal_out_snr = "snr_{}_{}".format(snr_min, snr_max)

    even_noise_filenames = {}
    even_signal_filenames = {}
    odd_noise_filenames = {}
    odd_signal_filenames = {}

    even_noise_save_outs = {}
    even_signal_save_outs = {}
    odd_noise_save_outs = {}
    odd_signal_save_outs = {}

    for dirname in save_options:
        even_noise_save_outs[dirname] = []
        even_signal_save_outs[dirname] = []
        odd_noise_save_outs[dirname] = []
        odd_signal_save_outs[dirname] = []

    skip_list = []#config["data"]["hardware_injections"]

    print(f"iterating over Nbands: {len(range_bands)}")
    # main loop over all sub bands
    for i in range(len(range_bands)):
        
        # set initial band frequency
        k = range_bands[i]
        # band width is 0.1 Hz
        width = 0.1

        # choose sub bands to skip, this list is hardware injections
        skip_band = False
        for lin in skip_list:
            if k > lin[0] and k < lin[1]:
                skip_band = True
                break

        if skip_band:
            continue

        #snr_start,snr_end = 80,150
        ms = int(range_index[i])
        me = int(ms+nbin*stride)

        # put odd bands into one folder and even bands into another folder
        if odd_even[i] == 1:
            #outpath = os.path.join(path,"odd")
            split = "odd"
        else:
            #outpath = os.path.join(path,"even")
            split = "even"
            
        # shift data up and down in frequency by n bins    
        shifts = [ms,ms + 30*stride,ms - 30*stride,ms + 60*stride, ms - 60*stride]

        # loop over all of the augmentation shifts
        for cts in shifts:
            #set band as index plut 180 bins i.e. 0.1 Hz
            cte = int(cts + 180*stride)
            if cts < 0 or cte >= sfts.H1.nbins:
                continue
            
            # get the median of the running medians as an estimate of the noise floow in this band.
            av_sh = [(2/tsft)*np.nanmedian(sfts.H1.rng_med[:,cts:cte],axis=1),(2/tsft)*np.nanmedian(sfts.L1.rng_med[:,cts:cte],axis=1)]
                
            # set lower frequency
            flow = k + cts / tsft
            
            # amke copy of data so dont overwrite sft
            datah = copy.deepcopy(sfts.H1.norm_sft_power[:,cts:cte])
            datal = copy.deepcopy(sfts.L1.norm_sft_power[:,cts:cte])
            
            # run injections function with new data, this saves noise iamge and injects signal and saves signal images
            noise_outs1, signal_outs1 = run_and_inj(datah,datal,flow,width,resize_image=resize_image,av_sh=av_sh,gen_noise_only=gen_noise_only,snrrange=snrrange, save_options=  save_options, stride = stride, degfree = degfree, brange = brange, sfts=sfts)
            
            del datah, datal
            
            # flip data in time
        
            datah,shH = flip_data_time(copy.deepcopy(sfts.H1.norm_sft_power[:,cts:cte]),sh=av_sh[0])
            datal,shL = flip_data_time(copy.deepcopy(sfts.L1.norm_sft_power[:,cts:cte]),sh=av_sh[1])
            
            noise_outs2, signal_outs2 = run_and_inj(datah,datal,flow,width,resize_image=resize_image,av_sh=[shH,shL],gen_noise_only=gen_noise_only,snrrange=snrrange, save_options=save_options, stride = stride, degfree = degfree, brange = brange,sfts=sfts)
        
            del datah,datal
        
            # flip data in freq
        
            datah = flip_data_freq(copy.deepcopy(sfts.H1.norm_sft_power[:,cts:cte]))
            datal = flip_data_freq(copy.deepcopy(sfts.L1.norm_sft_power[:,cts:cte]))
        
            noise_outs3, signal_outs3 = run_and_inj(datah,datal,flow,width,resize_image=resize_image,av_sh=av_sh,gen_noise_only=gen_noise_only,snrrange=snrrange, save_options=save_options, stride = stride, degfree = degfree, brange = brange,sfts=sfts)
            
            del datah,datal
            
            # roll data in time by 100 bins
        
            datah,shH = roll_in_time(copy.deepcopy(sfts.H1.norm_sft_power[:,cts:cte]),numbins=100,sh=av_sh[0])
            datal,shL = roll_in_time(copy.deepcopy(sfts.L1.norm_sft_power[:,cts:cte]),numbins=100,sh=av_sh[1])
        
            noise_outs4, signal_outs4 = run_and_inj(datah,datal,flow,width,resize_image=resize_image,av_sh=[shH,shL],gen_noise_only=gen_noise_only,snrrange=snrrange, save_options=save_options, stride = stride, degfree = degfree, brange = brange,sfts=sfts)
            
            del datah,datal

            for key in save_options:
                if split == "even":
                    even_noise_save_outs[key].extend([noise_outs1[key],noise_outs2[key],noise_outs3[key],noise_outs4[key] ])
                    even_signal_save_outs[key].extend([signal_outs1[key],signal_outs2[key],signal_outs3[key],signal_outs4[key]])
                elif split == "odd":
                    odd_noise_save_outs[key].extend([noise_outs1[key],noise_outs2[key],noise_outs3[key],noise_outs4[key] ])
                    odd_signal_save_outs[key].extend([signal_outs1[key],signal_outs2[key],signal_outs3[key],signal_outs4[key]])


    print("saving data to file .......")


    even_noise_filenames = os.path.join(*[path,"even",brange,noise_out_snr,f"freq_{bandmin}_{bandmax}_{len(shifts)*len(range_bands)}.hdf5"])
    even_signal_filenames = os.path.join(*[path,"even",brange,signal_out_snr,f"freq_{bandmin}_{bandmax}_{len(shifts)*len(range_bands)}.hdf5"])
    odd_noise_filenames = os.path.join(*[path,"odd",brange,noise_out_snr,f"freq_{bandmin}_{bandmax}_{len(shifts)*len(range_bands)}.hdf5"])
    odd_signal_filenames = os.path.join(*[path,"odd",brange,signal_out_snr,f"freq_{bandmin}_{bandmax}_{len(shifts)*len(range_bands)}.hdf5"])

    for fname, temp_data in zip([even_noise_filenames, even_signal_filenames, odd_noise_filenames, odd_signal_filenames], [even_noise_save_outs, even_signal_save_outs, odd_noise_save_outs, odd_signal_save_outs]):
        if not os.path.isdir(os.path.dirname(fname)):
            os.makedirs(os.path.dirname(fname))
       # print(fname, save_options)
        with h5py.File(fname,"w") as f:
            for key in save_options:
                if key == "pars":
                    # pars are stored as a dictionary so have to be saved differently to hdf5
                    try:
                        parkeys = list(temp_data[key][0].keys())
                    except:
                        print(np.shape(temp_data))
                        print("temp_datakeys", temp_data[key][0])
                        raise Exception(f"temp data has no data: {fname}")
                    f.create_dataset("parnames", data = parkeys)
                    f.create_dataset("pars", data = np.array([[td[key] for key in parkeys] for td in temp_data[key]]))
                else:
                    try:
                        f.create_dataset(key, shape=np.shape(temp_data[key]), data=np.array(temp_data[key]))
                    except ValueError as e:
                        print(key)
                        print(len(temp_data[key]))
                        lengths = [temp_data[key][i] for i in range(len(temp_data[key]))]
                        print(len(temp_data[key][0]))
                        print(max(lengths), min(lengths))
                        #print(np.shape(temp_data[key]))
                        #print(np.array(temp_data[key]).dtype)
                        #print(np.array(temp_data[key]).shape)
                        print(e)
                        raise Exception("Error on hdf5 save array")

    """
    for key in save_options:
        even_noise_filenames[key] = os.path.join(*[path,"even",key,brange,noise_out_snr,"freq_{}_{}.pkl".format(bandmin, len(even_noise_save_outs[key]))])
        even_signal_filenames[key] = os.path.join(*[path,"even",key,brange,signal_out_snr,"freq_{}_{}.pkl".format(bandmin, len(even_signal_save_outs[key]))])
        odd_noise_filenames[key] = os.path.join(*[path,"odd",key,brange,noise_out_snr,"freq_{}_{}.pkl".format(bandmin, len(odd_noise_save_outs[key]))])
        odd_signal_filenames[key] = os.path.join(*[path,"odd",key,brange,signal_out_snr,"freq_{}_{}.pkl".format(bandmin, len(odd_signal_save_outs[key]))])
    
        for fname in [even_noise_filenames[key], even_signal_filenames[key], odd_noise_filenames[key], odd_signal_filenames[key]]:
            if not os.path.isdir(os.path.dirname(fname)):
                os.makedirs(os.path.dirname(fname))

        with open(even_noise_filenames[key],"wb") as f:
            pickle.dump(even_noise_save_outs[key], f)
        with open(odd_noise_filenames[key],"wb") as f:
            pickle.dump(odd_noise_save_outs[key], f)
        with open(even_signal_filenames[key],"wb") as f:
            pickle.dump(even_signal_save_outs[key], f)
        with open(odd_signal_filenames[key],"wb") as f:
            pickle.dump(odd_signal_save_outs[key], f)
    """

    end_time = time.time()
    date = datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    with open(os.path.join(path,"timing.txt"),"a") as f:
        f.write("time-for-gen    {}    {} \n".format(end_time-start_time,date))




def run_and_inj(datah,datal,fmin,width,resize_image=False,snrrange = (50,150),av_sh=None,gen_noise_only=False, save_options = None, stride = 1,degfree = 96, brange = "", sfts = None):
    """_summary_

    Args:
        datah (_type_): input hanford detector data
        datal (_type_): input livingston detector data
        fmin (_type_): minimum frequency of band [Hz]
        width (_type_): band width [Hz]
        resize_image (bool, optional): _description_. Defaults to False.
        snrrange (tuple, optional): _description_. Defaults to (50,150).
        av_sh (_type_, optional): _description_. Defaults to None.
        gen_noise_only (bool, optional): _description_. Defaults to False.
        save_options (_type_, optional): _description_. Defaults to None.
        stride (int, optional): _description_. Defaults to 1.
        degfree (int, optional): _description_. Defaults to 96.
        brange (str, optional): _description_. Defaults to "".
        sfts (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """
    if np.shape(datah) != np.shape(datal):
        print("shapes do not match", "fmin ", fmin, stride )
        print("HL shape: ", np.shape(datah), np.shape(datal))
        return 0

    if av_sh is not None:
        Sn = {"H1":av_sh[0],"L1":av_sh[1]}
    # hardcoded start times as nsft
    nsft, tstart, tsft, flow, fhigh = sfts.H1.nsft, sfts.H1.epochs[0], 1800., fmin,fmin+width
    fmax = fmin + width
    # not gen_noise_only means that we also generate the noise bands as well as the injected bands
    if gen_noise_only: 
        # save noise realisation 1
        out_snr_noise = "snr_0.0_0.0"
        sig = cw.GenerateSignal()
        sig.alpha = 0
        sig.delta = 0
        sig.cosi = 0
        sig.phi0 = 0
        sig.psi = 0
        sig.f = [0,0]
        sig.earth_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/earth00-40-DE430.dat.gz"
        sig.sun_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/sun00-40-DE430.dat.gz"
        sig.snr = 0
        sig.fmin = fmin
        sig.fmax = fmax
        sig.tref = tstart
        
        Sn = {"H1":av_sh[0],"L1":av_sh[1]}
        data = sig.get_spectrogram(tstart = tstart, nsft = nsft,tref=tstart,tsft=tsft,fmin=fmin,fmax=fmax,snr=0,dets= ["H1","L1"],Sn=Sn)
        data.sum_sfts()
        data.downsamp_frequency(stride=stride,data_type="summed_norm_sft_power",remove_original=False)

        noise_outputs = cnn_data_gen.return_outputs(out_snr=out_snr_noise,datah=data.H1.downsamp_summed_norm_sft_power,datal=data.L1.downsamp_summed_norm_sft_power, fmin=fmin,fmax=fmax,snr=0,depth=0,h0=0, resize_image=resize_image, save_options = save_options, epochs = data.H1.summed_epochs, degfree = degfree, brange = brange)
        noise_outputs["pars"] = {"snr":0}
        noise_outputs["pars"]["tref"] = tstart
        noise_outputs["pars"]["snr"] = 0
        noise_outputs["pars"]["h0"] = 0
        noise_outputs["pars"]["depth"] = np.inf
        noise_outputs["pars"]["av_sh"] = np.nanmedian(av_sh)
        noise_outputs["pars"]["width"] = width
        noise_outputs["pars"]["fmin"] = data.H1.fmin
        noise_outputs["pars"]["fmax"] = data.H1.fmax
    # inject signal into band

    sig = cw.GenerateSignal()
    
    snrstart,snrend = snrrange[0],snrrange[1]
            
    sigfreq = np.random.uniform(0,1)*(width)*0.5 + flow + (width)*0.25
        
    snr = int(np.random.rand(1)*(snrend-snrstart) + snrstart)
        
    # places f0 in and around the frequency band and within the defined ranges for other parmaeters
    params_low =  [fmin + 0.025,      -1e-9 ,0      ,-1, 0      ,0      ,-1.0]
    params_high = [fmin + width-0.025, -1e-16,  2*np.pi, 1,2*np.pi,np.pi/2,1.0]
    data_list_signal = []
    data_list_noise = []
    nsft = int(nsft)
    param_list = ["f","fd","alpha","sindelta","phi0","psi","cosi"]
    pars = {}
    for j in np.linspace(0,len(param_list)-1,len(param_list)).astype(int):
        pars[param_list[j]] = params_low[j] + np.random.uniform(0,1)*(params_high[j] - params_low[j])

    #pars["fd"] = np.random.uniform(-1,1)*10**pars["fd"]
    #sunephem,earthephem = "/home/joseph.bayley/.virtualenvs/soap/share/lalpulsar/earth00-19-DE200.dat.gz","/home/joseph.bayley/.virtualenvs/soap/share/lalpulsar/sun00-19-DE200.dat.gz"

    sig.alpha = pars["alpha"]
    sig.delta = np.arcsin(pars["sindelta"])
    sig.cosi = pars["cosi"]
    sig.phi0 = pars["phi0"]
    sig.psi = pars["psi"]
    sig.f = [pars["f"],pars["fd"]]
    sig.earth_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/earth00-40-DE430.dat.gz"
    sig.sun_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/sun00-40-DE430.dat.gz"
    sig.snr = snr
    sig.fmin = fmin
    sig.fmax = fmax
    sig.tref = tstart

    #print("MEANS",flow,fhigh)
    #print(np.mean(datah,axis=1),np.mean(datal,axis=1))
            
    data = sig.get_spectrogram(tstart = tstart, nsft = len(datah),tref=tstart,tsft=tsft,fmin=flow,fmax=fhigh,snr=snr,noise_spect={"H1":datah,"L1":datal})
    data.sum_sfts()
    data.downsamp_frequency(stride=stride,data_type="summed_norm_sft_power",remove_original=False)

    epochs = data.H1.summed_epochs

    h0 = np.sqrt(np.nanmedian(av_sh))/data.depth

    inj_track = (sig.get_pulsar_path(epochs,"H1") - sig.fmin)*data.tsft

    pars["tref"] = tstart
    pars["snr"] = snr
    pars["av_sh"] = np.nanmedian(av_sh)
    pars["h0"] = h0
    pars["depth"] = data.depth
    pars["fmin"] = fmin
    pars["fmax"] = fmax

    out_snr_inj = "snr_{}_{}".format(snrstart,snrend)
    signal_outs = cnn_data_gen.return_outputs(out_snr=out_snr_inj,datah=data.H1.downsamp_summed_norm_sft_power,datal=data.L1.downsamp_summed_norm_sft_power, fmin=fmin,fmax=fmax,snr=snr, depth=data.depth,h0=h0,resize_image=resize_image,pars=pars,inj_track=inj_track,epochs=epochs, save_options=save_options, degfree = degfree, brange = brange)

    return noise_outputs, signal_outs




import unittest
import numpy as np

try:    
    from soapcw.line_aware_stat import gen_lookup
    cpp_implementation = True
except:
    cpp_implementation = False
    print("No cpp implementation available: install boost and gsl")

from soapcw.line_aware_stat import gen_lookup_python


class TestCWLoad(unittest.TestCase):
    """Tests for cw loading part of the SOAP package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.arr = np.linspace(0.1,100,5)
        self.las_py_approx = gen_lookup_python.LineAwareStatistic(self.arr, ndet=1 ,k=2, N=48, signal_prior_width=2, line_prior_width=3, noise_line_model_ratio=0.5, approx=True)
        self.las_py_exact = gen_lookup_python.LineAwareStatistic(self.arr, ndet=1 ,k=2, N=48, signal_prior_width=2, line_prior_width=3, noise_line_model_ratio=0.5, approx=False)
        if cpp_implementation:
            self.las_cpp_exact = gen_lookup.LineAwareStatistic(self.arr, ndet=1 ,k=2, N=48, signal_prior_width=2, line_prior_width=3, noise_line_model_ratio=0.5)

        self.arr2 = np.linspace(1,100,2)
        self.las_py_approx_2det = gen_lookup_python.LineAwareStatistic(self.arr2, ndet=2 ,k=2, N=48, signal_prior_width=2, line_prior_width=3, noise_line_model_ratio=0.5, approx=True)
        self.las_py_exact_2det = gen_lookup_python.LineAwareStatistic(self.arr2, ndet=2 ,k=2, N=48, signal_prior_width=2, line_prior_width=3, noise_line_model_ratio=0.5, approx=False)

        if cpp_implementation:
            self.las_cpp_exact = gen_lookup.LineAwareStatistic(self.arr2, ndet=1 ,k=2, N=48, signal_prior_width=2, line_prior_width=3, noise_line_model_ratio=0.5)


    def tearDown(self):
        """Tear down test fixtures, if any."""


    def test_gen_lookup_1det_py_approx(self):
        """Test the 1 detector lookup table generation approximate python version"""
        signoiseline = np.array([0.4180712556743122,  0.46573706068942955, 0.5238572889042366, 0.5944770691440852,  0.6772228852444641 ])

        np.testing.assert_almost_equal(signoiseline , self.las_py_approx.signoiseline)
    
    def test_gen_lookup_1det_py_exact(self):
        """Test the 1 detector lookup table generation approximate python version"""
        signoiseline = np.array([0.4168356127981346,  0.46470907883461643, 0.5230282098726995, 0.5938407592402759,  0.6767699501326778 ])

        np.testing.assert_almost_equal(signoiseline , self.las_py_exact.signoiseline)

    def test_gen_lookup_1det_cpp_exact(self):
        """Test the 1 detector lookup table generation approximate python version"""
        signoiseline = np.array([[0.4168403438629973,  0.46470910643499247, 0.5230282061301725, 0.5938407592400682,  0.6767699495907581 ] ])

        if cpp_implementation:
            np.testing.assert_almost_equal(signoiseline , self.las_cpp_exact.signoiseline)
    
    def test_gen_lookup_2det_py_approx(self):
        """Test the 1 detector lookup table generation approximate python version"""
        signoiseline = np.array([[0.28139104078142746, 0.3778010481689715 ],
                                    [0.3778010481689715,  0.6919634194611941 ]])

        np.testing.assert_almost_equal(signoiseline , self.las_py_approx_2det.signoiseline)
    
    def test_gen_lookup_2det_py_exact(self):
        """Test the 1 detector lookup table generation approximate python version"""
        signoiseline = np.array([[0.27942796216954896, 0.37667858291387774],
                                [0.37667858291387774, 0.6915352664549508 ]])

        np.testing.assert_almost_equal(signoiseline , self.las_py_exact_2det.signoiseline)




# -*- coding: utf-8 -*-
from __future__ import absolute_import
"""Top-level package for soapcw."""

__author__ = """Joe Bayley"""
__email__ = 'joseph.bayley@glasgow.ac.uk'
__version__ = '0.2.3'
from .soapcw import single_detector, two_detector, three_detector, single_detector_gaps
from .tools import tools, plots
from .cnn import __init__
from .neville import __init__
from .line_aware_stat import __init__
from .cw import __init__
from . import soap_config_parser

#try:
#    from .lookup_table import gen_lookup_python, save_lookup, gen_lookup
#except:
#    from .lookup_table import gen_lookup_python, save_lookup

